// File Name:
//		unload.c
//
// Contents:
//		Routines that perform driver cleanup
//

#include "xxdriver.h"

//
// Forward declarations of local routines
//
// (None)

//
// If the compiler can handle it, make the cleanup
// pageable so it won't be hanging around all the
// time
//
#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, XxDriverUnload )
#pragma alloc_text( page, XxReleaseHardware )
#endif // ALLOC_PRAGMA

//++
// Function:
//		XxDriverUnload
//
// Description:
//		This function releases any resources 
//		allocated during driver initialization.
//
// Arguments:
//		A pointer to the Driver object
//
// Return Value:
//		(None)
//--
VOID
XxDriverUnload(
	IN PDRIVER_OBJECT DriverObject
	)
{
    //
    // Stop interrupt processing and release hardware
    //
	XxReleaseHardware( DriverObject );
}

//++
// Function:
//		XxReleaseHardware
//
// Description:
//		This function releases any resources 
//		allocated during driver initialization.
//
// Arguments:
//		A pointer to the Driver object
//
// Return Value:
//		(None)
//--
VOID
XxReleaseHardware(
	IN PDRIVER_OBJECT DriverObject
	)
{
	PDEVICE_OBJECT pDevObj;
	PDEVICE_EXTENSION pDevExt;
	
	UNICODE_STRING linkName;
	WCHAR linkNameBuffer[ XX_MAX_NAME_LENGTH ];
	
	UNICODE_STRING	number;
	WCHAR numberBuffer[10];

	CM_RESOURCE_LIST ResList;
	BOOLEAN bConflict;

	linkName.Buffer = linkNameBuffer;
	linkName.MaximumLength = XX_MAX_NAME_LENGTH;

	number.Buffer = numberBuffer;
	number.MaximumLength = 10;
	
	
	pDevObj = DriverObject->DeviceObject;

	//
	// Traverse the list of Device objects
	// and clean up each one in turn...
	//	
	while( pDevObj != NULL ) {

		pDevExt = pDevObj->DeviceExtension;

		//
		// Add code here to save the state of
		// the hardware in the Registry and/or
		// to set the hardware into a known condition.
		//

		//
		// Stop handling interrupts from device
		//
		XxDisableInterrupts( pDevExt );

		IoDisconnectInterrupt( pDevExt->pInterrupt );
		
		//
		// Deallocate hardware resources belonging
		// only to this Device object...
		//
		ResList.Count = 0;	// Build an empty list
	
		IoReportResourceUsage(
			NULL,	    	// Default class name
			DriverObject,	// Ptr to Driver object
			NULL,			// No driver resources
			0,
			pDevObj,		// Ptr to Device object
			&ResList,		// Device resources
			sizeof( ResList ),
			FALSE,
			&bConflict );	// Junk, but required

		//
		// Form the Win32 symbolic link name.
	  	//
		linkName.Length = 0;
		RtlAppendUnicodeToString( 
			&linkName,
			XX_WIN32_DEVICE_NAME );

		//
		// Attach Win32 device number to the
		// end of the name; DOS device numbers
		// are one greater than NT numbers...
		//
		number.Length = 0;
		RtlIntegerToUnicodeString( 
			pDevExt->NtDeviceNumber + 1,
			10,
			&number ); 
		RtlAppendUnicodeStringToString( 
			&linkName,
			&number );
	    
		//        
		// Remove symbolic link from Object
		// namespace...
		//
		IoDeleteSymbolicLink( &linkName );

		//
		// Get address of next Device object
		// and get rid of the current one...
		//
		pDevObj = pDevObj->NextDevice;
		IoDeleteDevice( pDevExt->DeviceObject );
	}
	
	//
	// Deallocate hardware resources owned
	// by the Driver object...
	//
	ResList.Count = 0;	// Build an empty list
	
	IoReportResourceUsage(
		NULL,	    	// Default class name
		DriverObject,	// Pointer to Driver object
		&ResList,		// Driver resources
		sizeof( ResList ),
		pDevObj,		// Pointer to Device object
		NULL,			// Device resources
		0,
		FALSE,			// Don't override conflicts
		&bConflict );	// Junk, but required
}

