//++
// File Name:
//		transfer.c
//
// Contents:
//		Data transfer routines for parallel driver
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
static BOOLEAN
XxTransmitBytes( 
	IN PVOID Context
	);

//++
// Function:
//		XxStartIo
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual data transfer.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		(None)
//--
VOID
XxStartIo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );
		
	PDEVICE_EXTENSION pDE =
		DeviceObject->DeviceExtension;
		
	switch( IrpStack->MajorFunction ) {
		
		//
		// Use a SynchCritSection routine to
		// start the write operation...
		//
		case IRP_MJ_WRITE:
			//
			// Set up counts and byte pointer
			//
			pDE->BytesRequested = 
				IrpStack->Parameters.Write.Length;

			pDE->BytesRemaining = 
				pDE->BytesRequested;

			pDE->pBuffer =
				Irp->AssociatedIrp.SystemBuffer;

			//
			// Try to send the first bunch of
			// bytes. If there's a problem, use
			// the DPC routine to fail the IRP.
			//
			if( !KeSynchronizeExecution(
					pDE->pInterrupt,
					XxTransmitBytes,
					pDE ))
			{
				XxDpcForIsr(
					NULL,
					DeviceObject,
					Irp,
					pDE );
			}
			break;
		//
		// Should never get here -- just get rid
		// of the packet...
		//	
		default:
			Irp->IoStatus.Status =
						STATUS_NOT_SUPPORTED;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest(
				Irp,
				IO_NO_INCREMENT );
			IoStartNextPacket( DeviceObject, FALSE );
			break;
	}
}

//++
// Function:
//		XxTransmitBytes
//
// Description:
//		This function sends as many characters as
//		possible to device -- either one FIFO's worth
//		or as many as remain in the system buffer.
//
// Arguments:
//		Pointer to the Device Extension
//
// Return Value:
//		TRUE
//--
static BOOLEAN
XxTransmitBytes( 
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION pDE = 
		(PDEVICE_EXTENSION)Context;
	ULONG XferSize;

	UCHAR Control = XxReadControl( pDE );
	pDE->DeviceStatus = XxReadStatus( pDE );

	//
	// If all the bytes have been sent or the
	// device is unhappy, just quit
	//
	if( (pDE->BytesRemaining == 0)
		|| !XX_OK( pDE->DeviceStatus )) 
	{
		pDE->TransferInProgress = FALSE;
		return FALSE;
	}

	//
	// A transfer is happening. Calculate the number
	// of bytes to send in one bunch.
	//
	pDE->TransferInProgress = TRUE;

	if( pDE->BytesRemaining < pDE->FifoSize )
		XferSize = pDE->BytesRemaining;
	else
		XferSize = pDE->FifoSize;

	//
	// Send as many bytes to the device as it
	// can handle. Each byte must be strobed
	// out.
	//
	while( XferSize > 0 ) {

		//
		// Make sure the STROBE line is off
		//
		XxWriteControl( 
			pDE, 
			Control & ~XX_CTL_STROBE );

		//
		// Send a byte and hold it for at least
		// 500 nano-seconds
		//
		XxWriteData( pDE, *pDE->pBuffer );
		KeStallExecutionProcessor( 1 );

		//
		// Turn on the STROBE line and hold it
		// for at least 500 nano-seconds
		//
		XxWriteControl( 
			pDE, 
			Control | XX_CTL_STROBE );
		KeStallExecutionProcessor( 1 );

		//
		// Turn off the STROBE line
		//
		XxWriteControl( 
			pDE, 
			Control & ~XX_CTL_STROBE );
		KeStallExecutionProcessor( 1 );

		//
		// Update pointer and counters
		//
		pDE->pBuffer++;
		XferSize--;
		pDE->BytesRemaining--;
	}

	return TRUE;
}

//++
// Function:
//		XxIsr
//
// Description:
//		This function processes interrupts from
//		the parallel port
//
// Arguments:
//		Pointer to the Interrupt object
//		Pointer to the Device Extension 
//
// Return Value:
//		TRUE if we handled the interrupt
//--
BOOLEAN
XxIsr(
	IN PKINTERRUPT Interrupt,
	IN PVOID ServiceContext
	)
{
	PDEVICE_EXTENSION pDE = ServiceContext;
	PDEVICE_OBJECT pDevice = pDE->DeviceObject;
	PIRP Irp = pDevice->CurrentIrp;
	UCHAR Status = XxReadStatus( pDE );

	//
	// See if this device requested an interrupt
	//
	if(( Status & XX_STS_NOT_IRQ ) != 0 ) 
		return FALSE;
	
	//
	// Parallel port interrupted. If there's no
	// transfer in progress, just ignore it.
	// Otherwise, try to send the next bunch of
	// bytes. If XxTransmitBytes fails, it means
	// either an error occurred or there's no more
	// data to send.
	//
	if( pDE->TransferInProgress )
		if( !XxTransmitBytes( pDE ))
			IoRequestDpc( pDevice, Irp,	(PVOID)pDE );

	return TRUE;
}

//++
// Function:
//		XxDpcForIsr
//
// Description:
//		This function performs the low-IRQL
//		post-processing of I/O requests
//
// Arguments:
//		Pointer to a DPC object
//		Pointer to the Device object
//		Pointer to the IRP for this request
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxDpcForIsr(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION pDE = Context;
	
	Irp->IoStatus.Information =
			pDE->BytesRequested -
			pDE->BytesRemaining;

	//
	// Figure out what the final status
	// should be
	//
	if( XX_OK( pDE->DeviceStatus ))
		Irp->IoStatus.Status = STATUS_SUCCESS;

	else if( XX_POWERED_OFF( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_POWERED_OFF;

	else if( XX_NOT_CONNECTED( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_NOT_CONNECTED;

	else if( XX_OFF_LINE( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_OFF_LINE;

	else if( XX_PAPER_EMPTY( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_PAPER_EMPTY;

	else Irp->IoStatus.Status =	
			STATUS_DEVICE_DATA_ERROR;

	//
	// If we're being called directly from Start I/O,
	// don't give the user any priority boost.
	//
	if( Dpc == NULL )
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	else
		IoCompleteRequest( Irp, IO_PARALLEL_INCREMENT );

	//
	// This one's done. Begin working on the next
	//
	IoStartNextPacket( DeviceObject, FALSE );
}



	
	
	