//++
// File Name:
//		autocon.c
//
// Contents:
//		Routines that find auto-detected
//		hardware information in the Registry
//--

//
// Driver-specific header files
//
#include "xxdriver.h"

//
// Forward declarations of local routines
//
static NTSTATUS 
XxConfigCallback(
	IN PVOID Context,
	IN PUNICODE_STRING PathName,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PKEY_VALUE_FULL_INFORMATION *BusInfo,
	IN CONFIGURATION_TYPE CtrlrType,
	IN ULONG CtrlrNumber,
	IN PKEY_VALUE_FULL_INFORMATION *CtrlrInfo,
	IN CONFIGURATION_TYPE DeviceType,
	IN ULONG DeviceNumber,
	IN PKEY_VALUE_FULL_INFORMATION *DeviceInfo
	);

static BOOLEAN
XxGetPortInfo(
	IN PCM_PARTIAL_RESOURCE_DESCRIPTOR pPrd,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	);

static BOOLEAN
XxGetInterruptInfo(
	IN PCM_PARTIAL_RESOURCE_DESCRIPTOR pPrd,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	);

static NTSTATUS
XxFindMatchingConfigArray(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN OUT PCONFIG_ARRAY *ConfigList,
	OUT PCONFIG_ARRAY *ConfigArray
	);

//
// If the compiler can handle it, save space
// by making various routines discardable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( init, XxGetHardwareInfo )
#pragma alloc_text( init, XxConfigCallback )
#pragma alloc_text( init, XxGetPortInfo )
#pragma alloc_text( init, XxGetInterruptInfo )
#pragma alloc_text( init, XxFindMatchingConfigArray )
#endif

//++
// Function:
//		XxGetHardwareInfo
//
// Description:
//	This function extracts auto-detected information
//	about a device from the Registry.
//
// Arguments:
//		Registry path of driver's service key
//		Pointer to variable that receives address
//		  of list of Config Array blocks.
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_INSUFFICIENT_RESOURCES
//		STATUS_NO_SUCH_DEVICE
//--		
NTSTATUS
XxGetHardwareInfo(
	IN PUNICODE_STRING RegistryPath,	// (unused)
	OUT PCONFIG_ARRAY *ConfigList
	)
{
	INTERFACE_TYPE InterfaceType;
	ULONG InterfaceNumber;

	CONFIGURATION_TYPE CtrlrType = ParallelController;
	PCONFIG_ARRAY ConfigArray;

	NTSTATUS status;
	
	*ConfigList = NULL;	// No devices located yet

	//
	// Run through all the various bus types and
	// see if our device is on any of them...
	//
	for( InterfaceType = 0;
		 InterfaceType < MaximumInterfaceType;
		 InterfaceType++ ) {

		InterfaceNumber = 0;

		do {
			status = IoQueryDeviceDescription(
						&InterfaceType,
						&InterfaceNumber,
						&CtrlrType,
						NULL,
						NULL,
						NULL,
						XxConfigCallback,
						ConfigList );

			//
			// Return to caller if a real
			// error occurs
			//
			if( !NT_SUCCESS( status ) 
				&& status != 
				STATUS_OBJECT_NAME_NOT_FOUND )
			{
				XxReleaseHardwareInfo( *ConfigList );
				*ConfigList = NULL;
				return status;
			}

			InterfaceNumber++;

		} while( status != 
					STATUS_OBJECT_NAME_NOT_FOUND );

	} // end of for-loop

	if( *ConfigList == NULL )
		return STATUS_NO_SUCH_DEVICE;
	else
		return STATUS_SUCCESS;
}

//++
// Function:
//		XxConfigCallback
//
// Description:
//		This function is called once for each piece
//		of hardware that matches the pattern.
//
// Arguments:
//		(many -- see below)
//
// Return Value:
//		This function returns STATUS_SUCCESS if
//		it succeeds in adding the current device
//		to the Config block; otherwise some
//		error condition. 
//--
static NTSTATUS 
XxConfigCallback(
	IN PVOID Context,
	IN PUNICODE_STRING PathName,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PKEY_VALUE_FULL_INFORMATION *BusInfo,
	IN CONFIGURATION_TYPE CtrlrType,
	IN ULONG CtrlrNumber,
	IN PKEY_VALUE_FULL_INFORMATION *CtrlrInfo,
	IN CONFIGURATION_TYPE DeviceType,
	IN ULONG DeviceNumber,
	IN PKEY_VALUE_FULL_INFORMATION *DeviceInfo
	)
{
	//
	// So we don't have to typecast the context.
	//
    PCONFIG_ARRAY *ConfigList = Context;
	
	//
	// Short-hand pointers to resource data
	//
	PCM_FULL_RESOURCE_DESCRIPTOR pFrd;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR pPrd;
	PCONFIG_ARRAY ConfigArray;
	PDEVICE_BLOCK DeviceBlock;
	
	//
	// These booleans will tell us whether we got
	// all the information that we needed.
	//
	BOOLEAN bFoundPort = FALSE;
	BOOLEAN bFoundInterrupt = FALSE;

	NTSTATUS status;
	ULONG i;			// Generic loop control

	//
	// Locate the Config Array for this bus
	//
	status = XxFindMatchingConfigArray(
				BusType,
				BusNumber,
				ConfigList,
				&ConfigArray );
	if( !NT_SUCCESS( status ))
	{
		return status;
	}

	//
	// See if there's any room left in the Config
	// Array; if not, just drop this device on the
	// floor
	//
	if( ConfigArray->Count >= XX_MAXIMUM_DEVICES )
	{
		return STATUS_SUCCESS;
	}

	//
	// Make it easier to refer to the slot in the
	// Config Array belonging to this device
	//
	DeviceBlock = 
		&ConfigArray->Device[ConfigArray->Count];
	
	//
	// Get pointer to beginning of configuration
	// data for this device in the Registry
	//		
	pFrd = (PCM_FULL_RESOURCE_DESCRIPTOR)
		(((PUCHAR)CtrlrInfo
				[IoQueryDeviceConfigurationData])
		+ CtrlrInfo
				[IoQueryDeviceConfigurationData]
					->DataOffset);

	//
	// Loop through all Partial Resource Descriptors
	// looking for Port and Interrupt information
	//
	for ( i = 0;
		 i < pFrd->PartialResourceList.Count;
		 i++ )
	{
		pPrd = &pFrd->PartialResourceList
					.PartialDescriptors[i];

		//
		// Switch on the various partial resource
		// types. Pull out the pieces we need...
		//
		switch( pPrd->Type ) {

			case CmResourceTypePort: 

				bFoundPort = 
					XxGetPortInfo(
						pPrd, 
						BusType, 
						BusNumber, 
						DeviceBlock );
				break;

			case CmResourceTypeInterrupt:

				bFoundInterrupt =
					XxGetInterruptInfo(
						pPrd,
						BusType,
						BusNumber,
						DeviceBlock );
				break;

			default:
				break;
				
		} // end of switch
	} // end of for-loop
	
	//
	// Make sure we found all the pieces...
	//
	if( !( bFoundPort && bFoundInterrupt ))
	{
		return STATUS_NO_SUCH_DEVICE;
	}

	//
	// Account for the slot that we've just
	// filled up...
	//	
	ConfigArray->Count++;
	return STATUS_SUCCESS;
}

//++
// Function:
//		XxGetPortInfo
//
// Description:
//		This function pulls I/O Port infomation
//		from a Partial Resource Descriptor
//
// Arguments:
//		Pointer to a Partial Resource Descriptor
//		Bus type for this device
//		Bus number for this device
//		Pointer to this device's slot in Config Array
//
// Return Value:
//		This function returns TRUE if we found the
//		data we wanted, FALSE otherwise.
//--
static BOOLEAN
XxGetPortInfo(
	IN PCM_PARTIAL_RESOURCE_DESCRIPTOR pPrd,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK DeviceBlock
	)
{
	PHYSICAL_ADDRESS TranslatedPortBase;
	ULONG uAddressSpace = 1;
	
	DeviceBlock->OriginalPortBase = 
							pPrd->u.Port.Start;
				 
	DeviceBlock->PortSpan = pPrd->u.Port.Length;

	if( !HalTranslateBusAddress(
			BusType,
			BusNumber,
			DeviceBlock->OriginalPortBase,
			&uAddressSpace,
			&TranslatedPortBase	)) {
		return FALSE;
	}

	DeviceBlock->PortBase = 
		(PUCHAR)TranslatedPortBase.LowPart;

	return TRUE;	
}

//++
// Function:
//		XxGetInterruptInfo
//
// Description:
//		This function pulls Interrupt infomation
//		from a Partial Resource Descriptor
//
// Arguments:
//		Pointer to a Partial Resource Descriptor
//		Bus type for this device
//		Bus number for this device
//		Pointer to this device's slot in Config Array
//
// Return Value:
//		This function returns TRUE if we found the
//		data we wanted, FALSE otherwise.
//--
static BOOLEAN
XxGetInterruptInfo(
	IN PCM_PARTIAL_RESOURCE_DESCRIPTOR pPrd,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK DeviceBlock
	)
{
	if( pPrd->Flags == CM_RESOURCE_INTERRUPT_LATCHED )
		DeviceBlock->InterruptMode = Latched;
	else
		DeviceBlock->InterruptMode = LevelSensitive;

	DeviceBlock->OriginalIrql = 
			pPrd->u.Interrupt.Level;

	DeviceBlock->OriginalVector = 
			pPrd->u.Interrupt.Vector;

	DeviceBlock->ShareVector = FALSE;
	DeviceBlock->FloatingSave = FALSE;

	DeviceBlock->SystemVector = 
			HalGetInterruptVector(
				BusType,
 				BusNumber,
				pPrd->u.Interrupt.Level,
				pPrd->u.Interrupt.Vector,
				&DeviceBlock->Dirql,
				&DeviceBlock->Affinity );

	return TRUE;	
}

//++
// Function:
//		XxFindMatchingConfigArray
//
// Description:
//		This function finds or creates a Config Array
//		for a specific bus type and number.
//
// Arguments:
//		Bus type to match
//		Bus number to match
//		Pointer to variable containing 
//			root of config list
//		Pointer to variable receiving 
//			address of matchting Config Array
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_INSUFFICIENT_RESOURCES
//
//--
static NTSTATUS
XxFindMatchingConfigArray(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN OUT PCONFIG_ARRAY *ConfigList,
	OUT PCONFIG_ARRAY *ConfigArray
	)
{
	PCONFIG_ARRAY CurrentConfigArray = *ConfigList;

	//
	// Run the linked list of Config Arrays and
	// look for matching bus information.
	//
	while( CurrentConfigArray != NULL )
	{
		//
		// If this Config Array matches, return
		// it to the caller and quit.
		//
		if( CurrentConfigArray->
				BusType == BusType &&
			CurrentConfigArray->
				BusNumber == BusNumber )
		{
			*ConfigArray = CurrentConfigArray;
			return STATUS_SUCCESS;
		}

		//
		// Otherwise, get the next candidate from
		// the list and try again.
		//
		CurrentConfigArray = 
			CurrentConfigArray->NextConfigArray;
	}

	//
	// Either, the list is empty or there is no
	// matching Config Array. Create a new one...
	//
	if(( *ConfigArray = ExAllocatePool(
							PagedPool, 
							sizeof( CONFIG_ARRAY )))
		== NULL )
	{
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	RtlZeroMemory( 
		*ConfigArray, 
		sizeof( CONFIG_ARRAY ));

	//
	// Add the new Config Array to the head of
	// the caller's linked list.
	//
	(*ConfigArray)->NextConfigArray = *ConfigList;
	*ConfigList = *ConfigArray;

	//
	// Initialize various fields in the new
	// Config Array.
	//
	(*ConfigArray)->BusType = BusType;
	(*ConfigArray)->BusNumber = BusNumber;
	(*ConfigArray)->Count = 0;

	return STATUS_SUCCESS;
}
