//++
// File Name:
//		hardware.h
//
// Contents:
//		Hardware-specific stuff for parallel driver
//--
#ifndef __XX_HARDWARE__
#define __XX_HARDWARE__

//
// Define which types of hardware resources this
// device needs to allocate. Here, the device needs
// ports and interrupt resources, but not memory or
// DMA. If one device uses more than one of the same
// resource, (two disjoint ranges of I/O space, for 
// example) set the symbol equal to the number of
// items it uses.
//
#define	XX_PORT_RESOURCE_ITEMS			1
#define	XX_INTERRUPT_RESOURCE_ITEMS		1
#define	XX_MEMORY_RESOURCE_ITEMS		0
#define	XX_DMA_RESOURCE_ITEMS			0

//
// Define the number of CM_PARTIAL_RESOURCE_DESCRIPTOR
// items needed for each piece of hardware belonging
// to this driver.
//
#define XX_RESOURCE_ITEMS_PER_DEVICE		\
(	(XX_PORT_RESOURCE_ITEMS)			+	\
	(XX_INTERRUPT_RESOURCE_ITEMS)		+	\
	(XX_MEMORY_RESOURCE_ITEMS)			+	\
	(XX_DMA_RESOURCE_ITEMS)					)

//
// Default values for hardware
//
#define XX_DEF_PORT			0x0378
#define XX_DEF_PORT_SPAN	3
#define XX_DEF_IRQ			7
#define	XX_FIFO_SIZE		1

//
// Define device registers as relative offsets
//
#define XX_DATA			0 // Data register
#define XX_STATUS		1 // Status register
#define XX_CONTROL		2 // Control register

//
// Status register bits:
//
#define XX_STS_NOT_IRQ	0x04	// 0 Int requested
#define XX_STS_NOT_ERR	0x08	// 0 Error
#define XX_STS_SELCTD	0x10	// 1 Selected
#define XX_STS_PE		0x20	// 1 Paper empty
#define XX_STS_NOT_ACK	0x40	// 0 Acknowledge
#define XX_STS_NOT_BSY	0x80	// 0 Printer busy

//
// Control register bits:
//
#define XX_CTL_STROBE	0x01	// 1 Strobe data
#define XX_CTL_AUTOLF	0x02	// 1 Auto line feed
#define XX_CTL_NOT_INI	0x04	// 0 Init printer
#define XX_CTL_SELECT	0x08	// 1 Select printer
#define XX_CTL_INTENB	0x10	// 1 Interrupt enable
#define	XX_CTL_DEFAULT	0xC0	// (Sets unused bits)

//
// Define access macros for registers. Each macro takes
// a pointer to a Device Extension as an argument
//
#define XxReadStatus( pDevExt )				\
(READ_PORT_UCHAR(							\
  pDevExt->PortBase + XX_STATUS ))

#define XxReadControl( pDevExt )			\
(READ_PORT_UCHAR(							\
  pDevExt->PortBase + XX_CONTROL ))

#define XxWriteControl( pDevExt, bData )	\
(WRITE_PORT_UCHAR(							\
  pDevExt->PortBase + XX_CONTROL, bData ))

#define XxWriteData( pDevExt, bData )		\
(WRITE_PORT_UCHAR(							\
  pDevExt->PortBase + XX_DATA, bData ))

//
// The following macros check the printer status
// byte for various kinds of conditions.
//

#define XX_POWERED_OFF( Status ) (					\
((Status & XX_STS_NOT_ERR) ^ XX_STS_NOT_ERR) &&		\
((Status & XX_STS_NOT_ACK) ^ XX_STS_NOT_ACK) && 	\
( Status & XX_STS_NOT_BSY) )

#define XX_POWERED_ON(Status) (						\
((Status & XX_STS_NOT_BSY) ^ XX_STS_NOT_BSY) &&		\
( Status & XX_STS_SELCTD) &&						\
( Status & XX_STS_NOT_ERR) )

#define XX_OFF_LINE( Status ) (						\
( Status & XX_STS_NOT_ERR) &&						\
((Status & XX_STS_NOT_BSY) ^ XX_STS_NOT_BSY) &&		\
 !(Status & XX_STS_SELCTD) )

#define XX_ONLINE(Status) (							\
( Status & XX_STS_NOT_ERR) &&						\
( Status & XX_STS_NOT_BSY) &&						\
((Status & XX_STS_PE) ^ XX_STS_PE) &&				\
( Status & XX_STS_SELCTD) )

#define XX_SELECTED(Status) (						\
( Status & XX_STS_SELCTD) )

#define XX_OK(Status) (								\
( Status & XX_STS_NOT_ERR) &&						\
((Status & XX_STS_PE) ^ XX_STS_PE) &&				\
( Status & XX_STS_NOT_BSY) )


#define XX_BUSY(Status) (							\
((Status & XX_STS_NOT_BSY)	^ XX_STS_NOT_BSY) &&	\
( Status & XX_STS_NOT_ERR) )

#define XX_PAPER_EMPTY( Status ) (					\
( Status & XX_STS_PE ) )

#define XX_NOT_CONNECTED( Status ) (				\
( Status & XX_STS_NOT_ERR) &&						\
( Status & XX_STS_NOT_BSY) &&						\
!(Status & XX_STS_SELCTD) )

#define XX_NO_CABLE(Status) (						\
((Status & XX_STS_NOT_BSY) ^ XX_STS_NOT_BSY) &&		\
( Status & XX_STS_NOT_ACK) &&						\
( Status & XX_STS_PE) &&							\
( Status & XX_STS_SELCTD) &&						\
( Status & XX_STS_NOT_ERR))

#endif // __XX_HARDWARE__

