//++
// File Name:
//		init.c
//
// Contents:
//		Driver initialization code
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
static NTSTATUS
XxCreateDevice (	
	IN PDRIVER_OBJECT DriverObject,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK DeviceBlock,
	IN ULONG NtDeviceNumber
	);

static BOOLEAN
XxInitDevice(
	IN PVOID SynchContext
	);

//
// If the compiler can handle it, make the DriverEntry
// routine discardable, so that it doesn't waste space
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( init, DriverEntry )
#pragma alloc_text( init, XxCreateDevice )
#pragma alloc_text( init, XxInitDevice )
#endif // ALLOC_PRAGMA

//++
// Function:
//		DriverEntry
//
// Description:
//		This function initializes the driver, locates
//		and claims hardware resources, and creates
//		various NT objects needed to process I/O
//		requests.
//
// Arguments:
//		Pointer to the Driver object
//		Registry path string for driver service key
//
// Return Value:
//		NTSTATUS signaling success or failure
//--
NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	)
{
	PCONFIG_ARRAY ConfigList;
	PCONFIG_ARRAY ConfigArray;
	ULONG NtDeviceNumber;
	NTSTATUS status;
	ULONG i;

	//
	// Load up the Config list...
	//
	status = XxGetHardwareInfo(
				RegistryPath,
				&ConfigList );
	if( !NT_SUCCESS( status ))
	{
		return status;
    }
	
	//
	// Allocate the hardware...
	//
	status = XxReportHardwareUsage(
				DriverObject,
				ConfigList );
	if( !NT_SUCCESS( status ))
	{
		XxReleaseHardwareInfo( ConfigList );
		return status;
	}

	//
	// Export other driver entry points...
	//
	DriverObject->DriverStartIo = XxStartIo;
 	DriverObject->DriverUnload = XxDriverUnload;
	
	DriverObject->MajorFunction[ IRP_MJ_CREATE ] =
		XxDispatchOpenClose;
	DriverObject->MajorFunction[ IRP_MJ_CLOSE ] =
		XxDispatchOpenClose;
	DriverObject->MajorFunction[ IRP_MJ_WRITE ] =
		XxDispatchWrite;

	//
	// Initialize a Device object for each piece
	// of hardware we've found
	//
	ConfigArray = ConfigList;
	NtDeviceNumber = 0;

	while( ConfigArray != NULL )
	{
		for( i = 0;
			 i < ConfigArray->Count;
			 i++ )
		{
   			status = XxCreateDevice(
   						DriverObject,
						ConfigArray->BusType,
						ConfigArray->BusNumber,
						&ConfigArray->Device[i],
						NtDeviceNumber );

   			if( !NT_SUCCESS( status )) break;
			NtDeviceNumber++;
		}

		if( !NT_SUCCESS( status )) break;
		
		//
		// Get next array in the chain
		//
		ConfigArray = ConfigArray->NextConfigArray;
	}

	if( !NT_SUCCESS( status ))
	{
		XxReleaseHardware( DriverObject );
	}

	XxReleaseHardwareInfo( ConfigList );

	return status;
}

//++
// Function:
//		XxCreateDevice
//
// Description:
//		This function creates and initializes a
//		single Device object, using a set of hardware
//		descriptions from the Config list.
//
// Arguments:
//		Pointer to the Driver object
//		Type of bus to which device is attached
//		Number of bus where device is attached
//		Pointer to config data about the device
//		Number to associate with NT Device object
//
// Return Value:
//		NTSTATUS signaling success or failure
//--
static NTSTATUS
XxCreateDevice (
	IN PDRIVER_OBJECT DriverObject,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK DeviceBlock,
	IN ULONG NtDeviceNumber
	)
{
	NTSTATUS status;
	
	PDEVICE_OBJECT pDevObj;
    PDEVICE_EXTENSION pDevExt;

	UNICODE_STRING deviceName;
	WCHAR deviceNameBuffer[ XX_MAX_NAME_LENGTH ];
	
	UNICODE_STRING linkName;
	WCHAR linkNameBuffer[ XX_MAX_NAME_LENGTH ];
	
	UNICODE_STRING	number;
	WCHAR numberBuffer[10];

	number.Buffer = numberBuffer;
	number.MaximumLength = 10;
	
	//
   	// Form the base NT device name...
   	//
	deviceName.Buffer = deviceNameBuffer;
	deviceName.MaximumLength = XX_MAX_NAME_LENGTH;
	deviceName.Length = 0;
   	RtlAppendUnicodeToString(
		&deviceName,
   		XX_NT_DEVICE_NAME );

	//
	// Convert the device number into a string and
	// attach it to the end of the device name.
	//
	number.Length = 0;
	RtlIntegerToUnicodeString( 
		NtDeviceNumber, 
		10, 
		&number ); 
	RtlAppendUnicodeStringToString( 
		&deviceName,
		&number );

    //
    // Create a Device object for this device...
    //
	status = IoCreateDevice(
				DriverObject,
				sizeof( DEVICE_EXTENSION ),
				&deviceName,
				FILE_DEVICE_PARALLEL_PORT,
				0,
				TRUE,
				&pDevObj );

	if( !NT_SUCCESS( status ))
	{
		return status;
    }

	pDevObj->Flags |= DO_BUFFERED_IO;

	//
    // Initialize the device extension structure
    //
	pDevExt = pDevObj->DeviceExtension;
    pDevExt->DeviceObject = pDevObj;
	pDevExt->NtDeviceNumber = NtDeviceNumber;
	pDevExt->FifoSize = XX_FIFO_SIZE;
	
	//
	// Copy things from Device Block
	//
	pDevExt->PortBase = DeviceBlock->PortBase;
	
	//
	// Calculate the polling interval
	//
	pDevExt->PollingInterval =
		RtlConvertLongToLargeInteger(
			XX_POLLING_INTERVAL * 10 * -1 );

	//
	// Prepare the polling timer and its DPC object
	//
	KeInitializeTimer( &pDevExt->PollingTimer );

	KeInitializeDpc(
		&pDevExt->PollingDpc,
		XxPollingTimerDpc,
		(PVOID)pDevObj );

	//
   	// Form the Win32 symbolic link name.
   	//
	linkName.Buffer = linkNameBuffer;
	linkName.MaximumLength = XX_MAX_NAME_LENGTH;
	linkName.Length = 0;
	RtlAppendUnicodeToString( 
		&linkName,
		XX_WIN32_DEVICE_NAME );
	//
	// Reset the number string and do another
	// conversion. Win32 device numbers are
	// one greater than the NT equivalent.
	//
	number.Length = 0;
	RtlIntegerToUnicodeString( 
		NtDeviceNumber + 1,
		10,
		&number ); 
	RtlAppendUnicodeStringToString( 
		&linkName,
		&number );
	    
	//        
	// Create a symbolic link so our device is
	// visible to Win32...
	//
 	status = IoCreateSymbolicLink(
 				&linkName,
				&deviceName );
    
	//
    // See if the symbolic link was created...
    //
	if( !NT_SUCCESS( status )) 
	{
    	IoDeleteDevice( pDevObj );
		return status;
    }

	//
	// Initialize the hardware
	//
	XxInitDevice( pDevExt );

	return status;
}

//++
// Function:
//		XxInitDevice
//
// Description:
//		This function cycles the INIT line to
//		initialize the hardware. It also sets
//		the SELECT line.
//
// Arguments:
//		Pointer to a Device Extension
//
// Return Value:
//		BOOLEAN TRUE
//--
static BOOLEAN
XxInitDevice(
	IN PVOID SynchContext
	)
{
	PDEVICE_EXTENSION pDE =
		(PDEVICE_EXTENSION) SynchContext;

	//
	// Make sure device interrupts are OFF and
	// the INIT cycle isn't in progress
	//
	XxWriteControl(
		pDE,  
		XX_CTL_DEFAULT 
		| XX_CTL_NOT_INI );
	KeStallExecutionProcessor( 60 );

	//
	// Clear the INIT bit to start a device
	// initialization cycle
	//
	XxWriteControl( pDE, XX_CTL_DEFAULT );
	KeStallExecutionProcessor( 60 );

	//
	// Set the INIT bit to complete the cycle,
	// select the device.
	//
	XxWriteControl( 
		pDE, 
		XX_CTL_DEFAULT
		| XX_CTL_NOT_INI
		| XX_CTL_SELECT );
	KeStallExecutionProcessor( 60 );

	return TRUE;
}




	