//++
// File Name:
//		transfer.c
//
// Contents:
//		Data transfer routines for parallel driver
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
static BOOLEAN
XxTransmitBytes( 
	IN PDEVICE_EXTENSION pDE
	);

static VOID
XxFinishCurrentRequest(
	IN PDEVICE_OBJECT DeviceObject,
	IN PDEVICE_EXTENSION pDE,
	IN PIRP Irp,
	IN CCHAR PriorityBoost
	);

//++
// Function:
//		XxStartIo
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual data transfer.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		(None)
//--
VOID
XxStartIo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );
		
	PDEVICE_EXTENSION pDE =
		DeviceObject->DeviceExtension;
		
	switch( IrpStack->MajorFunction ) {
		
		case IRP_MJ_WRITE:
			//
			// Set up counts and byte pointer
			//
			pDE->BytesRequested = 
				IrpStack->Parameters.Write.Length;

			pDE->BytesRemaining = 
				pDE->BytesRequested;

			pDE->pBuffer =
				Irp->AssociatedIrp.SystemBuffer;

			//
			// Try to send the first bunch of
			// bytes. If there's a problem, 
			// fail the IRP.
			//
			if( !XxTransmitBytes( pDE ))
			{
				XxFinishCurrentRequest(
					DeviceObject,
					pDE,
					Irp,
					IO_NO_INCREMENT );
			}
			break;
		//
		// Should never get here -- just get rid
		// of the packet...
		//	
		default:
			Irp->IoStatus.Status =
						STATUS_NOT_SUPPORTED;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest(
				Irp,
				IO_NO_INCREMENT );
			IoStartNextPacket( DeviceObject, FALSE );
			break;
	}
}

//++
// Function:
//		XxTransmitBytes
//
// Description:
//		This function sends as many characters as
//		possible to device -- either one FIFO's worth
//		or as many as remain in the system buffer.
//
// Arguments:
//		Pointer to the Device Extension
//
// Return Value:
//		TRUE
//--
static BOOLEAN
XxTransmitBytes( 
	IN PDEVICE_EXTENSION pDE
	)
{
	ULONG XferSize;

	UCHAR Control = XxReadControl( pDE );
	pDE->DeviceStatus = XxReadStatus( pDE );

	//
	// If all the bytes have been sent or the
	// device is unhappy, just quit
	//
	if( (pDE->BytesRemaining == 0)
		|| !XX_OK( pDE->DeviceStatus )) 
	{
		return FALSE;
	}

	//
	// Calculate the number of bytes to 
	// send in one bunch.
	//
	if( pDE->BytesRemaining < pDE->FifoSize )
		XferSize = pDE->BytesRemaining;
	else
		XferSize = pDE->FifoSize;

	//
	// Send as many bytes to the device as it
	// can handle. Each byte must be strobed
	// out.
	//
	while( XferSize > 0 ) {

		//
		// Make sure the STROBE line is off
		//
		XxWriteControl( 
			pDE, 
			Control & ~XX_CTL_STROBE );

		//
		// Send a byte and hold it for at least
		// 500 nano-seconds
		//
		XxWriteData( pDE, *pDE->pBuffer );
		KeStallExecutionProcessor( 1 );

		//
		// Turn on the STROBE line and hold it
		// for at least 500 nano-seconds
		//
		XxWriteControl( 
			pDE, 
			Control | XX_CTL_STROBE );
		KeStallExecutionProcessor( 1 );

		//
		// Turn off the STROBE line
		//
		XxWriteControl( 
			pDE, 
			Control & ~XX_CTL_STROBE );
		KeStallExecutionProcessor( 1 );

		//
		// Update pointer and counters
		//
		pDE->pBuffer++;
		XferSize--;
		pDE->BytesRemaining--;
	}

	//
	// Start the polling timer
	//
	KeSetTimer(
		&pDE->PollingTimer,
		pDE->PollingInterval,
		&pDE->PollingDpc );

	return TRUE;
}

//++
// Function:
//		XxPollingTimerDpc
//
// Description:
//		This function sends data to the parallel
//		port in response to polling timer events
//
// Arguments:
//		Pointer to the Dpc object (Not used)
//		Pointer to the Device object
//		SystemArgument1 (Not Used)
//		SystemArgument2 (Not Used)
//
// Return Value:
//		(None)
//--
VOID
XxPollingTimerDpc(
	IN PKDPC Dpc,
	IN PVOID Context,
	IN PVOID SystemArgument1,
	IN PVOID SystemArgument2
	)
{
	PDEVICE_OBJECT DeviceObject = Context;

	//
	// Try to send the next bunch of bytes. If
	// XxTransmitBytes fails, it means either an
	// error occurred or there's no more data to
	// send.
	//
	if( !XxTransmitBytes( 
			DeviceObject->DeviceExtension ))
	{
		XxFinishCurrentRequest( 
			DeviceObject, 
			DeviceObject->DeviceExtension,
			DeviceObject->CurrentIrp,
			IO_PARALLEL_INCREMENT );
	}
}

//++
// Function:
//		XxFinishCurrentRequest
//
// Description:
//		This function performs all the 
//		post-processing of I/O requests
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the Device Extension
//		Pointer to the IRP for this request
//		Priority boost to be applied to Irp
//
// Return Value:
//		(None)
//--
static VOID
XxFinishCurrentRequest(
	IN PDEVICE_OBJECT DeviceObject,
	IN PDEVICE_EXTENSION pDE,
	IN PIRP Irp,
	IN CCHAR PriorityBoost
	)
{
	Irp->IoStatus.Information =
			pDE->BytesRequested -
			pDE->BytesRemaining;

	//
	// Figure out what the final status
	// should be
	//
	if( XX_OK( pDE->DeviceStatus ))
		Irp->IoStatus.Status = STATUS_SUCCESS;

	else if( XX_POWERED_OFF( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_POWERED_OFF;

	else if( XX_NOT_CONNECTED( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_NOT_CONNECTED;

	else if( XX_OFF_LINE( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_OFF_LINE;

	else if( XX_PAPER_EMPTY( pDE->DeviceStatus ))
		Irp->IoStatus.Status = 
			STATUS_DEVICE_PAPER_EMPTY;

	else Irp->IoStatus.Status =	
			STATUS_DEVICE_DATA_ERROR;

	//
	// If we're being called directly from Start I/O,
	// don't give the user any priority boost.
	//
	IoCompleteRequest( Irp, PriorityBoost );

	//
	// This one's done. Begin working on the next
	//
	IoStartNextPacket( DeviceObject, FALSE );
}



	
	
	