//++
// File Name:
//		dispatch.c
//
// Contents:
//		Dispatch routines for parallel driver
//--

//
// Driver-specific header files
//
#include "xxdriver.h"

//
// Forward declarations of local routines
//
// (None)

//
// If the compiler can handle it, save space
// by making various routines pageable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, XxDispatchOpen )
#pragma alloc_text( page, XxDispatchClose )
#pragma alloc_text( page, XxDispatchWrite )
#endif //ALLOC_PRAGMA

//++
// Function:
//		XxDispatchOpen
//
// Description:
//		This function dispatches CreateFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchOpen(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	//
	// Start the I/O time-out clock. If this were
	// a shared device, it would be necessary to
	// handle this in a more sophisticated way.
	//
	((PDEVICE_EXTENSION)
		pDO->DeviceExtension)->TimeRemaining = -1;

	IoStartTimer( pDO );

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

//++
// Function:
//		XxDispatchClose
//
// Description:
//		This function dispatches CloseHandle
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchClose(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	//
	// Last handle is being closed. Stop the
	// time-out clock.
	//
	IoStopTimer( pDO );

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

//++
// Function:
//		XxDispatchWrite
//
// Description:
//		This function dispatches WriteFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS 
XxDispatchWrite(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION IrpStack = 
				IoGetCurrentIrpStackLocation( Irp );
	//
	// Check for zero-length transfers
	//
	if( IrpStack->Parameters.Write.Length == 0 ) {
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	
	//
	// Start device operation
	//
	IoMarkIrpPending( Irp );
	IoStartPacket( pDO, Irp, 0, NULL );
	return STATUS_PENDING;
}


