//++
// File Name:
//		cancel.c
//
// Contents:
//		Support for IRP cancellation
//
//--

//
// Driver-specific header files
//
#include "xxdriver.h"

//
// Forward declarations of local routines
//
// (None)

//
// If the compiler can handle it, save space
// by making various routines pageable
//    
#ifdef ALLOC_PRAGMA
// (None)
#endif //ALLOC_PRAGMA

//++
// Function:
//		XxCleanupDeviceQueue
//
// Description:
//		This function removes any IRPs associated
//		with a given File object from a specific
//		Device Queue. It then cancels these IRPs.
//
//		This function is called from PASSIVE_LEVEL
//		but it spends time at DISPATCH_LEVEL while
//		it holds the Cancel spin lock.
//
// Arguments:
//		Pointer to Device Queue object
//		Pointer to File object
//
// Return Value:
//		(None)
//--
VOID
XxCleanupDeviceQueue(
	IN PKDEVICE_QUEUE DeviceQueue,
	IN PFILE_OBJECT FileObject
	)
{
	KIRQL OldIrql;
	PIRP CancelIrp;
	PIRP RequeueIrp;
 
	PIO_STACK_LOCATION CancelIrpStack;

	LIST_ENTRY CancelList;
	LIST_ENTRY RequeueList;
	PLIST_ENTRY ListHead;
	PKDEVICE_QUEUE_ENTRY QueueEntry;
	
	//
	// Prepare the two work lists
	//
	InitializeListHead( &CancelList );
	InitializeListHead( &RequeueList );

	//
	// This work must be done under the protection
	// of the Cancel spin lock...
	//
	IoAcquireCancelSpinLock( &OldIrql );

	//
	// See if there's actually any work
	// to do. If not, quit right now.
	//
	if( IsListEmpty( &DeviceQueue->DeviceListHead ))
	{
		IoReleaseCancelSpinLock( OldIrql );
		return;
	}

	//
	// Loop until every IRP has been removed from
	// the Device Queue. Decide whether each one
	// should be canceled or not. (At the end of
	// this loop, the Device Queue will have been
	// emptied; hence its state will be Not Busy.)
	//
	while(( QueueEntry = 
				KeRemoveDeviceQueue(
							DeviceQueue ))
									!= NULL )
	{
		CancelIrp = 
			CONTAINING_RECORD(
				QueueEntry,
				IRP,
				Tail.Overlay.DeviceQueueEntry );

		KdPrint(( "\tRemoving IRP %8x from Device Queue\n", CancelIrp ));

		CancelIrpStack = 
			IoGetCurrentIrpStackLocation( CancelIrp );

		//
		// If the IRP's File object is the same as
		// the one in the IRP_MJ_CLEANUP IRP, set 
		// the IRP's CancelXxx fields and put the 
		// IRP in a holding queue of IRPs to be 
		// completed.
		//
		if( CancelIrpStack->FileObject == FileObject )
		{
			CancelIrp->Cancel		 = TRUE;
			CancelIrp->CancelIrql	 = OldIrql;
			CancelIrp->CancelRoutine = NULL;

			KdPrint(( "\tAdding IRP %8x to CancelList\n", CancelIrp ));

			InsertTailList(
				&CancelList,
				&CancelIrp->Tail.Overlay.ListEntry );
		}

		//
		// Else, put it in a holding queue of IRP's
		// to be requeued... 
		//
		else
		{
			KdPrint(( "\tAdding IRP %8x to RequeueList\n", CancelIrp ));

			InsertTailList(
				&RequeueList,
				&CancelIrp->Tail.Overlay.ListEntry );
		}
	}

	//
	// Run the holding queue of IRP's to be requeued
	// and put them back in the Device Queue object
	//
	while( !IsListEmpty( &RequeueList ))
	{
		ListHead = RemoveHeadList( &RequeueList );
		RequeueIrp = 
			CONTAINING_RECORD(
				ListHead,
				IRP,
				Tail.Overlay.ListEntry );

		//
		// Requeue the IRP. If this is the first
		// reinsertion, the Device Queue will be
		// Not Busy. Hence, it will take two calls 
		// to get the insertion to work.
		//

		KdPrint(( "\tRequeuing IRP %8x to Device Queue\n", RequeueIrp ));

		if( !KeInsertDeviceQueue(
				DeviceQueue,
				&RequeueIrp->
					Tail.Overlay.DeviceQueueEntry ))
		{
			KeInsertDeviceQueue(
				DeviceQueue,
				&RequeueIrp->
					Tail.Overlay.DeviceQueueEntry );
		}
	}

	//
	// Then release the Cancel spin lock
	//
	IoReleaseCancelSpinLock( OldIrql );

	//
	// Run the length of the holding queue and 
	// complete every IRP that we found in it.
	//
	while( !IsListEmpty( &CancelList ))
	{
		ListHead = RemoveHeadList( &CancelList );
		CancelIrp = 
			CONTAINING_RECORD(
				ListHead,
				IRP,
				Tail.Overlay.ListEntry );

		KdPrint(( "\tCancelling IRP %8x to Device Queue\n", CancelIrp ));
		
		CancelIrp->IoStatus.Status = STATUS_CANCELLED;
		CancelIrp->IoStatus.Information = 0;

		IoCompleteRequest( 
			CancelIrp, 
			IO_NO_INCREMENT );
	}
}


//++
// Function:
//		XxCancelPrimaryIrp
//
// Description:
//		This function gets called to cancel
//		an IRP that's somewhere between the
//		primary Device Queue and the beginning
//		of processing in the Start I/O routine.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the target IRP
//
// Return Value:
//		(None)
//
// Synchronization:
//		This routine is called at DISPATCH_LEVEL
//		IRQL, holding the Cancel spin lock.
//
//--
VOID
XxCancelPrimaryIrp(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PDEVICE_EXTENSION DeviceExtension =
				DeviceObject->DeviceExtension;

	KdPrint(( "XxCancelPrimaryIrp: IRP %8x\n", Irp ));

	//
	// If IRP is already in the CurrentIrp slot,
	// but not yet started, we can cancel it.
	//
	if( Irp == DeviceObject->CurrentIrp )
	{
		KdPrint(( "\tIRP is current\n" ));

		//
		// Release the Cancel spin lock
		// *BEFORE* calling IoCompleteRequest
		//
		IoReleaseCancelSpinLock( Irp->CancelIrql );

		Irp->IoStatus.Status = STATUS_CANCELLED;
		Irp->IoStatus.Information = 0;

		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		IoStartNextPacket( DeviceObject, TRUE );
	}
	//
	// IRP is not the current IRP. It must still 
	// be sitting in the Device Queue.
	//
	else
	{
		KdPrint(( "\tIRP is in Device Queue; removing and cancelling.\n" ));
		KeRemoveEntryDeviceQueue( 
			&DeviceObject->DeviceQueue,
			&Irp->Tail.Overlay.DeviceQueueEntry );

		IoReleaseCancelSpinLock( Irp->CancelIrql );

		Irp->IoStatus.Status = STATUS_CANCELLED;
		Irp->IoStatus.Information = 0;

		//
		// Complete this IRP, but don't start the
		// next one.
		//
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	}
		
}


//++
// Function:
//		XxCancelAlternateIrp
//
// Description:
//		This function gets called to cancel
//		an IRP that's somewhere between the
//		alternate Device Queue and the beginning
//		of processing in the alternate Start I/O
//		routine.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the target IRP
//
// Return Value:
//		(None)
//
// Synchronization:
//		This routine is called at DISPATCH_LEVEL
//		IRQL, holding the Cancel spin lock.
//
//--
VOID
XxCancelAlternateIrp(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PDEVICE_EXTENSION DeviceExtension =
				DeviceObject->DeviceExtension;

	KdPrint(( "XxCancelAlternateIrp: IRP %8x\n", Irp ));

	//
	// If IRP is already in the CurrentIrp slot,
	// but not yet started, we can cancel it.
	//
	if( Irp == DeviceExtension->CurrentAlternateIrp )
	{
		KdPrint(( "\tIRP is current\n" ));

		//
		// Release the Cancel spin lock
		// *BEFORE* calling IoCompleteRequest
		//
		IoReleaseCancelSpinLock( Irp->CancelIrql );

		Irp->IoStatus.Status = STATUS_CANCELLED;
		Irp->IoStatus.Information = 0;

		IoCompleteRequest( Irp, IO_NO_INCREMENT );

		XxAlternateStartNextPacket( 
							DeviceObject,
							TRUE );
	}
	//
	// IRP is not the current IRP. It must still 
	// be sitting in the Device Queue.
	//
	else
	{
		KdPrint(( "\tIRP is in Device Queue; removing and cancelling.\n" ));

		KeRemoveEntryDeviceQueue( 
			&DeviceExtension->AlternateIrpQueue,
			&Irp->Tail.Overlay.DeviceQueueEntry );

		IoReleaseCancelSpinLock( Irp->CancelIrql );

		Irp->IoStatus.Status = STATUS_CANCELLED;
		Irp->IoStatus.Information = 0;

		//
		// Complete this IRP, but don't start the
		// next one.
		//
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	}
}

