//++
// File Name:
//		dispatch.c
//
// Contents:
//		Dispatch routines for UART driver
//--

//
// Driver-specific header files
//
#include "xxdriver.h"

//
// Forward declarations of local routines
//
// (None)

//
// If the compiler can handle it, save space
// by making various routines pageable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, XxDispatchOpenClose )
#pragma alloc_text( page, XxDispatchWrite )
#pragma alloc_text( page, XxDispatchRead )
#pragma alloc_text( page, XxDispatchCleanup )
#endif //ALLOC_PRAGMA

//++
// Function:
//		XxDispatchOpenClose
//
// Description:
//		This function dispatches CreateFile and
//		CloseHandle requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchOpenClose(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

//++
// Function:
//		XxDispatchWrite
//
// Description:
//		This function dispatches WriteFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS 
XxDispatchWrite(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION IrpStack = 
				IoGetCurrentIrpStackLocation( Irp );
	//
	// Check for zero-length transfers
	//
	if( IrpStack->Parameters.Write.Length == 0 )
	{
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	
	//
	// Start device operation
	//
	IoMarkIrpPending( Irp );
	
	IoStartPacket( 
			DeviceObject, 
			Irp, 
			0, 
			XxCancelPrimaryIrp );

	return STATUS_PENDING;
}

//++
// Function:
//		XxDispatchRead
//
// Description:
//		This function dispatches ReadFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS 
XxDispatchRead(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION IrpStack = 
				IoGetCurrentIrpStackLocation( Irp );
	//
	// Check for zero-length transfers
	//
	if( IrpStack->Parameters.Read.Length == 0 )
	{
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	
	//
	// Start device operation
	//
	IoMarkIrpPending( Irp );

	XxAlternateStartPacket( 
					DeviceObject, 
					Irp, 
					XxCancelAlternateIrp );

	return STATUS_PENDING;
}

//++
// Function:
//		XxDispatchCleanup
//
// Description:
//		This function processes 
//		IRP_MJ_CLEANUP requests
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchCleanup(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION CleanupIrpStack = 
			IoGetCurrentIrpStackLocation( Irp );

	PDEVICE_EXTENSION DeviceExtension = 
			DeviceObject->DeviceExtension;

	KdPrint(( "XxDispatchCleanup: IRP %8x\n", Irp ));

	//
	// Remove any pending IRP_MJ_WRITE IRPs 
	// from the primary Device Queue.
	//
	XxCleanupDeviceQueue(
		&DeviceObject->DeviceQueue,
		CleanupIrpStack->FileObject );

	//
	// Remove any pending IRP_MJ_READ IRPs
	// from the alternate Device Queue.
	//
	XxCleanupDeviceQueue(
		&DeviceExtension->AlternateIrpQueue,
		CleanupIrpStack->FileObject );

	//
	// Complete the IRP_MJ_CLEANUP IRP itself...
	//
	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}


