//++
// File Name:
//		hardware.h
//
// Contents:
//		Hardware-specific stuff for UART driver
//--
#ifndef __XX_HARDWARE__
#define __XX_HARDWARE__

//
// Define which types of hardware resources this
// device needs to allocate. Here, the device needs
// ports and interrupt resources, but not memory or
// DMA. If one device uses more than one of the same
// resource, (two disjoint ranges of I/O space, for 
// example) set the symbol equal to the number of
// items it uses.
//
#define	XX_PORT_RESOURCE_ITEMS			1
#define	XX_INTERRUPT_RESOURCE_ITEMS		1
#define	XX_MEMORY_RESOURCE_ITEMS		0
#define	XX_DMA_RESOURCE_ITEMS			0

//
// Define the number of CM_PARTIAL_RESOURCE_DESCRIPTOR
// items needed for each piece of hardware belonging
// to this driver.
//
#define XX_RESOURCE_ITEMS_PER_DEVICE		\
(	(XX_PORT_RESOURCE_ITEMS)			+	\
	(XX_INTERRUPT_RESOURCE_ITEMS)		+	\
	(XX_MEMORY_RESOURCE_ITEMS)			+	\
	(XX_DMA_RESOURCE_ITEMS)					)

//
// Default values for hardware
//
#define	XX_OUTPUT_FIFO_SIZE	16

//
// Indexes into a table of valid baud-rate
// divisors
//
#define	XX_BAUD_RATE_300		0
#define	XX_BAUD_RATE_600		1
#define	XX_BAUD_RATE_1200		2
#define	XX_BAUD_RATE_2400		3
#define	XX_BAUD_RATE_4800		4
#define	XX_BAUD_RATE_9600		5
#define	XX_BAUD_RATE_19200		6
#define XX_HIGHEST_BAUD_RATE	6

//
// Define device registers as relative offsets
//
#define XX_RCV_DATA		0 // Received data buff (R/O)
#define XX_TXMT_DATA	0 // Transmit data buff (W/O)
#define XX_BAUD_LSB		0 // LSB of baud-rate 

#define XX_INT_ENABLE	1 // Interrupt enable (R/W)
#define XX_BAUD_MSB		1 // MSB of baud-rate (R/W)

#define XX_INT_ID		2 // Interrupt ID (R/O)
#define XX_FIFO_CTRL	2 // FIFO control (W/O)

#define XX_LINE_CTRL	3 // Line control (R/W)

#define XX_MODEM_CTRL	4 // Modem control (R/W)

#define XX_LINE_STS		5 // Line status (R/O)

#define XX_MODEM_STS	6 // Modem status (R/O)

#define XX_SCRATCH		7 // Scratch (R/W -- unused)

//
// Interrupt Enable Register bits
//
#define XX_IEN_RDR		0x01
#define XX_IEN_TBE		0x02
#define XX_IEN_ERR		0x04
#define XX_IEN_RS232	0x08

//
// Interrupt ID Register bits
//
#define XX_IIR_ERR		0x06
#define XX_IIR_RDA		0x04
#define XX_IIR_FIFO_TMO	0x0c
#define XX_IIR_TBE		0x02
#define XX_IIR_RS232	0x00

#define XX_IIR_FIFOS_ENABLED_MASK	0xc0
#define XX_IIR_INTERRUPT_ID_MASK	0x0f
#define XX_IIR_NO_INTERRUPT			0x01

//
// FIFO Control Register bits
//
#define XX_FCR_ENABLE		((UCHAR)0x01)
#define XX_FCR_RCVR_RESET	((UCHAR)0x02)
#define XX_FCR_TXMT_RESET	((UCHAR)0x04)

#define XX_FCR_TRIGGER_1	((UCHAR)0x00)
#define XX_FCR_TRIGGER_4	((UCHAR)0x40)
#define XX_FCR_TRIGGER_8	((UCHAR)0x80)
#define XX_FCR_TRIGGER_14	((UCHAR)0xc0)

//
// Line Control Register bits
//
#define XX_LCR_DATA_5		((UCHAR)0x00)
#define XX_LCR_DATA_6		((UCHAR)0x01)
#define XX_LCR_DATA_7		((UCHAR)0x02)
#define XX_LCR_DATA_8		((UCHAR)0x03)
#define XX_LCR_DATA_MASK    ((UCHAR)0x03)

#define XX_LCR_STOP_1		((UCHAR)0x00)
#define XX_LCR_STOP_1_5		((UCHAR)0x04)
#define XX_LCR_STOP_2		((UCHAR)0x04)
#define XX_LCR_STOP_MASK	((UCHAR)0x04)

#define XX_LCR_PARITY_NONE	((UCHAR)0x00)
#define XX_LCR_PARITY_ODD	((UCHAR)0x08)
#define XX_LCR_PARITY_EVEN	((UCHAR)0x18)
#define XX_LCR_PARITY_MARK	((UCHAR)0x28)
#define XX_LCR_PARITY_SPACE	((UCHAR)0x38)
#define XX_LCR_PARITY_MASK	((UCHAR)0x38)

#define XX_LCR_BREAK		0x40
#define XX_LCR_DLAB			0x80

//
// Modem Control Register bits
//
#define XX_MCR_DTR			0x01
#define XX_MCR_RTS			0x02
#define XX_MCR_OUT1			0x04
#define XX_MCR_OUT2			0x08
#define XX_MCR_LOOP			0x10

//
// Line Status Register bits
//
#define XX_LSR_DATA_RDY			0x01
#define XX_LSR_RCVR_OVERRUN		0x02
#define XX_LSR_BAD_PARITY		0x04
#define XX_LSR_FRAME_ERR		0x08
#define XX_LSR_BREAK			0x10
#define XX_LSR_TXMT_BUFF_EMPTY	0x20
#define XX_LSR_SHIFT_REG_EMPTY	0x40
#define XX_LSR_FIFO_ERR			0x80

//
// Modem Status Register bits
//
#define XX_MSR_DELTA_CTS	0x01
#define XX_MSR_DELTA_DSR	0x02
#define XX_MSR_DELTA_RI		0x04
#define XX_MSR_DELTA_DCD	0x08
#define XX_MSR_CTS			0x10
#define XX_MSR_DSR			0x20
#define XX_MSR_RI			0x40
#define XX_MSR_DCD			0x80



//
// Define access macros for registers. Each macro
// takes a pointer to a Device Extension
//
#define XxReadDataBuffer( pDevExt )				\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_RCV_DATA ))

#define XxWriteDataBuffer( pDevExt, bData )		\
(WRITE_PORT_UCHAR(								\
  pDevExt->PortBase + XX_TXMT_DATA, bData ))

#define XxReadIntEnable( pDevExt )				\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_INT_ENABLE ))

#define XxWriteIntEnable( pDevExt, bData )		\
(WRITE_PORT_UCHAR(								\
  pDevExt->PortBase + XX_INT_ENABLE, bData ))

#define XxReadIntId( pDevExt )					\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_INT_ID ))

#define XxWriteFifoControl( pDevExt, bData )	\
(WRITE_PORT_UCHAR(								\
  pDevExt->PortBase + XX_FIFO_CTRL, bData ))

#define XxReadLineControl( pDevExt )			\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_LINE_CTRL ))

#define XxWriteLineControl( pDevExt, bData )	\
(WRITE_PORT_UCHAR(								\
  pDevExt->PortBase + XX_LINE_CTRL, bData ))

#define XxReadModemControl( pDevExt )			\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_MODEM_CTRL ))

#define XxWriteModemControl( pDevExt, bData )	\
(WRITE_PORT_UCHAR(								\
  pDevExt->PortBase + XX_MODEM_CTRL, bData ))

#define XxReadLineStatus( pDevExt )				\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_LINE_STS ))

#define XxReadModemStatus( pDevExt )			\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_MODEM_STS ))

#define XxReadBaudMsb( pDevExt )				\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_BAUD_MSB ))

#define XxWriteBaudMsb( pDevExt, bData )		\
(WRITE_PORT_UCHAR(								\
  pDevExt->PortBase + XX_BAUD_MSB, bData ))

#define XxReadBaudLsb( pDevExt )				\
(READ_PORT_UCHAR(								\
  pDevExt->PortBase + XX_BAUD_LSB ))

#define XxWriteBaudLsb( pDevExt, bData )		\
(WRITE_PORT_UCHAR(								\
  pDevExt->PortBase + XX_BAUD_LSB, bData ))

#endif // __XX_HARDWARE__

