//++
// File Name:
//		output.c
//
// Contents:
//		Routines to process UART output requests
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
// (None)

//++
// Function:
//		XxStartOutput
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual output operation.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		(None)
//--
VOID
XxStartOutput(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	KIRQL OldIrql;

	PDEVICE_EXTENSION DeviceExtension =
				DeviceObject->DeviceExtension;

	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );
		
 	//
	// Before starting the operation, see if the
	// Cancel routine has run between the time
	// the IRP was removed from the Device Queue
	// and now. Start by grabbing the Cancel spin
	// lock.
	//
	IoAcquireCancelSpinLock( &OldIrql );

	//
	// Check the state of the Cancel flag. If 
	// it's set, release the spin lock and return
	// immediately. The IRP has already been  
	// processed by the Cancel routine.
	//
	if( Irp->Cancel )
	{
		IoReleaseCancelSpinLock( OldIrql );
		return;
	}
	//
	// If the Cancel flag is clear, remove the
	// IRP from the cancelable state and start
	// processing it.
	//
	else
	{
		IoSetCancelRoutine( Irp, NULL );
		IoReleaseCancelSpinLock( OldIrql );
	}
		

	switch( IrpStack->MajorFunction ) {
		
		case IRP_MJ_WRITE:
			//
			// Set up counts and byte pointer
			//
			DeviceExtension->OutputBytesRequested = 
					IrpStack->Parameters.Write.Length;

			DeviceExtension->OutputBytesRemaining = 
				DeviceExtension->OutputBytesRequested;

			DeviceExtension->pOutputBuffer =
				Irp->AssociatedIrp.SystemBuffer;

			//
			// Try to send the first bunch of
			// bytes. If there's a problem, use
			// the DPC routine to fail the IRP.
			//
			if( !KeSynchronizeExecution(
					DeviceExtension->pInterrupt,
					XxTransmitBytes,
					DeviceExtension ))
			{
				XxDpcForOutputs(
					NULL,
					DeviceObject,
					Irp,
					DeviceExtension );
			}
			break;

		//
		// Should never get here -- just get rid
		// of the packet...
		//	
		default:

			//
			// Fail the IRP and start the next one.
			//
			Irp->IoStatus.Status = 
							STATUS_NOT_SUPPORTED;
			Irp->IoStatus.Information = 0;

			IoCompleteRequest( Irp, IO_NO_INCREMENT );
			IoStartNextPacket( DeviceObject, TRUE );
			break;
	}
}

//++
// Function:
//		XxTransmitBytes
//
// Description:
//		This function sends as many characters as
//		possible to device -- either one FIFO's worth
//		or as many as remain in the system buffer.
//
// Arguments:
//		Pointer to the Device Extension
//
// Return Value:
//		TRUE
//
// Synchronization:
//		This routine expects to be synchronized 
//		with the ISR, running at DIRQL holding 
//		the Interrupt spin lock.
//
//--
BOOLEAN
XxTransmitBytes( 
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION DeviceExtension = Context;
	ULONG XferSize;
	ULONG i;

	//
	// If all the bytes have been sent,
	// just quit
	//
	if( DeviceExtension->OutputBytesRemaining <= 0 )
	{
		DeviceExtension->
			OutputInterruptsValid = FALSE;

		return FALSE;
	}

	//
	// A transfer is happening. Calculate the number
	// of bytes to send in one bunch.
	//
	if( DeviceExtension->OutputBytesRemaining < 
					DeviceExtension->OutputFifoSize )
	{
		XferSize = 
			DeviceExtension->OutputBytesRemaining;
	}
	else
	{
		XferSize = DeviceExtension->OutputFifoSize;
	}

	//
	// Send as many bytes to the device as it
	// can handle. 
	//
	for( i=0; i<XferSize; i++ )
	{
		//
		// Send a byte 
		//
		XxWriteDataBuffer( 
			DeviceExtension, 
			*DeviceExtension->pOutputBuffer );

		//
		// Update pointer and counters
		//
		DeviceExtension->pOutputBuffer++;
	}

	DeviceExtension->OutputBytesRemaining -= XferSize;

	//
	// Interrupts are believable now
	//
	DeviceExtension->OutputInterruptsValid = TRUE;

	return TRUE;
}

//++
// Function:
//		XxDpcForOutputs
//
// Description:
//		This function performs the low-IRQL
//		post-processing of output requests
//
// Arguments:
//		Pointer to a DPC object
//		Pointer to the Device object
//		Pointer to the IRP for this request
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxDpcForOutputs(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION DeviceExtension = Context;

	//
	// Determine number of bytes transferred
	//
	Irp->IoStatus.Information =
			DeviceExtension->OutputBytesRequested -
			DeviceExtension->OutputBytesRemaining;

	//
	// Figure out what the final status
	// should be
	//
	Irp->IoStatus.Status = STATUS_SUCCESS;


	//
	// If we're being called directly from Start I/O,
	// don't give the user any priority boost.
	//
	if( Dpc == NULL )
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	else
		IoCompleteRequest( Irp, IO_SERIAL_INCREMENT );

	//
	// This one's done. Begin working on the next
	//
	IoStartNextPacket( DeviceObject, TRUE );
}



	
	
	