// File Name:
//		xxdriver.h
//
// Contents:
//		Constants, structures, and function
//		declarations specific to this driver.
//
#ifndef __XX_DRIVER__
#define __XX_DRIVER__

//
// Header files
//
#include <ntddk.h>
#include "hardware.h"

//
// Various constants
//
#define	XX_NT_DEVICE_NAME		L"\\Device\\Xx"
#define	XX_WIN32_DEVICE_NAME	L"\\DosDevices\\XX"
#define	XX_DRIVER_NAME			L"XXDRIVER"

#define	XX_MAXIMUM_DEVICES	10
#define	XX_MAX_NAME_LENGTH	50

//++
// Description:
//		Driver-defined structure used to hold 
//		miscellaneous device information.
//
// Access:
//		Allocated from NON-PAGED POOL
//		Available at any IRQL
//--
typedef struct _DEVICE_EXTENSION {
	PDEVICE_OBJECT DeviceObject; // Back pointer
	
	ULONG NtDeviceNumber;	// Zero-based device num	

	PUCHAR PortBase;		// First control register
	PKINTERRUPT pInterrupt;	// Interrupt object

	//
	// Current UART settings
	//
	ULONG InputFifoTriggerLevel;
	ULONG DataBits;
	ULONG StopBits;
	ULONG Parity;

	//
	// The Device Queue and DPC objects are used for
	// processing input requests.
	//
	KDEVICE_QUEUE AlternateIrpQueue;
	KDPC AlternateDpc;
	PIRP CurrentAlternateIrp;

	//
	// The following items keep track of our
	// progress during an output operation
	//
	ULONG OutputFifoSize;		// Bytes to send at once
	ULONG OutputBytesRequested;	// Total size
	ULONG OutputBytesRemaining;	// Chars left to send
	PUCHAR pOutputBuffer;		// Next char to send

	BOOLEAN OutputInterruptsValid;

	//
	// The following items keep track of our
	// progress during an input operation
	//
	ULONG InputFifoSize;		// Count of bytes
	ULONG InputBytesRequested;	// Input buffer size 
	ULONG InputBytesRemaining;	// Space left in buffer
	PUCHAR pInputBuffer;		// Next available slot

	BOOLEAN InputInterruptsValid;

	UCHAR	DeviceStatus;	// Most recent status
	
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;


//++
// Description:
//		Temporary struct used during driver 
//		initialization to hold config data about
//		a single device.
//
// Access:
//		Allocated from PAGED POOL
//		Available ONLY at IRQL PASSIVE_LEVEL
//--
typedef struct _DEVICE_BLOCK {

	//
	// Original values pulled from the Registry
	//
	PHYSICAL_ADDRESS OriginalPortBase;
	ULONG PortSpan;	

	ULONG OriginalIrql;
	ULONG OriginalVector;
	KINTERRUPT_MODE InterruptMode;
	BOOLEAN ShareVector;
	BOOLEAN FloatingSave;
	
	ULONG OriginalDmaChannel;

	//
	// Converted values that will be used by
	// the driver		

	PUCHAR PortBase;	// First control register

	ULONG SystemVector;
	KIRQL Dirql;
	KAFFINITY Affinity;

} DEVICE_BLOCK, *PDEVICE_BLOCK;

//++
// Description:
//		Temporary struct used during driver 
//		initialization to hold hardware 
//		configuration data about all devices.
//
// Access:
//		Allocated from PAGED POOL
//		Available ONLY at IRQL PASSIVE_LEVEL
//--
typedef struct _CONFIG_ARRAY
{
	//
	// We keep a list of these arrays, one
	// for each bus-type/bus-number combination
	// where we find our hardware.
	//
	struct _CONFIG_ARRAY *NextConfigArray;

	//
	// The bus to which all the devices are
	// attached.
	//
	INTERFACE_TYPE BusType;
	ULONG BusNumber;

	//
	// Number of devices in the array
	//
	ULONG	Count;
	
	//
	// One array-element for each device
	//
	DEVICE_BLOCK Device[ XX_MAXIMUM_DEVICES ];
	
} CONFIG_ARRAY, *PCONFIG_ARRAY;

//++
// Description:
//		Driver-defined structure used to change 
//		the baud-rate of a serial port. A structure
//		is necessary because there's only one context
//		argument available for the Synch Crit Section
//		routine that actually makes the change.
//
// Access:
//		Allocated from NON-PAGED POOL
//		Available at any IRQL
//
//--
typedef struct _NEW_BAUD_RATE {
	PDEVICE_EXTENSION	DeviceExtension;
	ULONG Speed;
} NEW_BAUD_RATE, *PNEW_BAUD_RATE;

//
// Prototypes for globally defined functions...
//
// init.c
//
NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	);

//
// unload.c
//
VOID
XxDriverUnload(
	IN PDRIVER_OBJECT DriverObject
	);

VOID
XxReleaseHardware(
	IN PDRIVER_OBJECT DriverObject
	);

//
// regcon.c
//
NTSTATUS
XxGetHardwareInfo(
	IN PUNICODE_STRING RegistryPath,
	OUT PCONFIG_ARRAY *ConfigList
	);

//
// resalloc.c
//
NTSTATUS
XxReportHardwareUsage(
	IN PDRIVER_OBJECT DriverObject,
	IN PCONFIG_ARRAY pConfig
	);

VOID
XxReleaseHardwareInfo(
	IN PCONFIG_ARRAY ConfigList
	);

//
// dispatch.c
//
NTSTATUS
XxDispatchOpenClose(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

NTSTATUS 
XxDispatchWrite(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

NTSTATUS 
XxDispatchRead(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

NTSTATUS 
XxDispatchCleanup(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

//
// devqueue.c
//
VOID
XxAlternateStartPacket(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PDRIVER_CANCEL CancelFunction
	);

VOID
XxAlternateStartNextPacket(
	IN PDEVICE_OBJECT DeviceObject,
	IN BOOLEAN Cancelable
	);

//
// output.c
//
VOID
XxStartOutput(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

BOOLEAN
XxTransmitBytes( 
	IN PVOID Context
	);

VOID
XxDpcForOutputs(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	);

//
// input.c
//
VOID
XxAlternateStartIo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

BOOLEAN
XxReceiveBytes( 
	IN PVOID Context
	);

VOID
XxDpcForInputs(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	);

//
// isr.c
//
BOOLEAN
XxIsr(
	IN PKINTERRUPT Interrupt,
	IN PVOID ServiceContext
	);

//
// hardware.c
//
BOOLEAN
XxSetBaudRate(
	PNEW_BAUD_RATE NewBaudRate
	);

BOOLEAN
XxInitializeHardware(
	PDEVICE_EXTENSION pDE
	);

BOOLEAN
XxEnableInterrupts(
	PDEVICE_EXTENSION pDE
	);

BOOLEAN
XxDisableInterrupts(
	PDEVICE_EXTENSION pDE
	);

//
// cancel.c
//
VOID
XxCleanupDeviceQueue(
	IN PKDEVICE_QUEUE DeviceQueue,
	IN PFILE_OBJECT FileObject
	);

VOID
XxCancelPrimaryIrp(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

VOID
XxCancelAlternateIrp(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

#endif	// __XX_DRIVER__

