//++
// File Name:
//		resalloc.c
//
// Contents:
//		Routines that show how to allocate
//		a driver's hardware resources
//--

//
// Driver-specific header files
//
#include "xxdriver.h"

//
// Forward declarations of local routines
//
static PCM_PARTIAL_RESOURCE_DESCRIPTOR
XxBuildPartialDescriptors(
	IN PDEVICE_BLOCK Device,
	IN PCM_PARTIAL_RESOURCE_DESCRIPTOR Prd
	);

//
// If the platform can handle it, save space
// by making various routines discardable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( init, XxReportHardwareUsage )
#pragma alloc_text( init, XxBuildPartialDescriptors )
#pragma alloc_text( init, XxReleaseHardwareInfo )
#endif

//++
// Function:
//		XxReportHardwareUsage
//
// Description:
//		This function allocates any hardware used
//		by the driver
//
// Arguments:
//		Pointer to the Driver object
//		Pointer to linked-list of Config arrays,
//			one for each bus-type/bus-number
//			combination containing our hardware
//
// Return Value:
//		This function returns STATUS_SUCCESS, or
//		STATUS_INSUFFICIENT_RESOURCES.
//
//--
NTSTATUS
XxReportHardwareUsage(
	IN PDRIVER_OBJECT DriverObject,
	IN PCONFIG_ARRAY ConfigList
	)
{
	ULONG ListSize;
	PCM_RESOURCE_LIST ResourceList;
	PCM_FULL_RESOURCE_DESCRIPTOR Frd;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR Prd;
	PCONFIG_ARRAY CurrentArray;
	
	BOOLEAN bConflictDetected;
	NTSTATUS status;
	ULONG i;

	//
	// Calculate size of resource list
	//
	ListSize = 
		FIELD_OFFSET( CM_RESOURCE_LIST, List[0] );

	CurrentArray = ConfigList;

	while( CurrentArray != NULL )
	{
		ListSize +=
			sizeof( CM_FULL_RESOURCE_DESCRIPTOR ) +
			((( CurrentArray->Count * 
				XX_RESOURCE_ITEMS_PER_DEVICE ) - 1 ) *
				sizeof( 
					CM_PARTIAL_RESOURCE_DESCRIPTOR ));

		CurrentArray = CurrentArray->NextConfigArray;
	}

	//
	// Try and allocate paged memory for the resource
	// list. If it works, zero out the list.
	//
	ResourceList = 
		ExAllocatePool( PagedPool, ListSize );
	if( ResourceList == NULL )
	{
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	
	RtlZeroMemory( ResourceList, ListSize );
	
	//
	// Run the linked-list of Config arrays and
	// build a CM_FULL_RESOURCE_DESCRIPTOR for
	// each one.
	//
	CurrentArray = ConfigList;
	Frd = &ResourceList->List[0];

	while( CurrentArray != NULL )
	{
		ResourceList->Count++;
		Frd->InterfaceType = CurrentArray->BusType;
		Frd->BusNumber = CurrentArray->BusNumber;

		//
		// Set the number of Partial Resource
		// Descriptors in this FRD.
		//
		Frd->
			PartialResourceList.Count =
				CurrentArray->Count * 
					XX_RESOURCE_ITEMS_PER_DEVICE;

		//
		// Get pointer to first Partial Resource
		// Descriptor in this FRD.
		//
		Prd = &Frd->
				PartialResourceList.
					PartialDescriptors[0];

		//
		// Loop through all Device slots in the
		// current Config array. Build Partial
		// Resource Descriptors for the items
		// used by each device in the array.
		//
		for( i=0; i < CurrentArray->Count; i++ )
		{
			Prd = XxBuildPartialDescriptors(
					&CurrentArray->Device[i],
					Prd );
		}

		//
		// Point to beginning of next Full Resource
		// Descriptor.
		//
		(PUCHAR)Frd +=
			((( Frd->PartialResourceList.Count - 1 ) *
				sizeof( 
				  CM_PARTIAL_RESOURCE_DESCRIPTOR ))  +
			sizeof( CM_FULL_RESOURCE_DESCRIPTOR ));

		//
		// Get next Config array from linked-list
		//
		CurrentArray = CurrentArray->NextConfigArray;
	}

	//
	// Request ownership; resources will belong to
	// the Driver object.
	//
	status = IoReportResourceUsage(
				NULL,	// Optional class name
				DriverObject,
				ResourceList,
				ListSize,
				NULL,
				NULL,
				0,
				FALSE,	// Don't override
				&bConflictDetected );

	ExFreePool( ResourceList );

	if( !NT_SUCCESS( status ) || bConflictDetected )
		return STATUS_INSUFFICIENT_RESOURCES;
	else
		return STATUS_SUCCESS; 
}

//++
// Function:
//		XxBuildPartialDescriptors
//
// Description:
//		This function adds all the appropriate PRDs
//		for one piece of hardware to the current 
//		FRD. 
//
// Arguments:
//		Pointer to a Device Block describing the hardware
//		Pointer to the first available PRD to be used
//			for this device.
//
// Return Value:
//		This function returns a pointer to the
//		next free PRD in the current FRD.
//
//--
static PCM_PARTIAL_RESOURCE_DESCRIPTOR
XxBuildPartialDescriptors(
	IN PDEVICE_BLOCK Device,
	IN PCM_PARTIAL_RESOURCE_DESCRIPTOR Prd
	)
{
	//
	// Set up PRD for control registers
	//
	Prd->Type = CmResourceTypePort;
	
	Prd->ShareDisposition = 
			CmResourceShareDriverExclusive;
			
	Prd->Flags = CM_RESOURCE_PORT_IO;
	
	Prd->u.Port.Start = 
				Device->OriginalPortBase;
	Prd->u.Port.Length = Device->PortSpan;
	Prd++;		

	//
	// Set up PRD for Interrupt resource
	//
	Prd->Type = CmResourceTypeInterrupt;
	
	Prd->ShareDisposition = 
			CmResourceShareDriverExclusive;

	if( Device->InterruptMode == Latched )
	  Prd->Flags = 
		CM_RESOURCE_INTERRUPT_LATCHED;
	else
	  Prd->Flags = 
		CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE;

	Prd->u.Interrupt.Level =
				Device->OriginalIrql;

	Prd->u.Interrupt.Vector = 
				Device->OriginalVector;
	Prd++;

	//
	// Set up PRD for DMA resource
	//
	Prd->Type = CmResourceTypeDma;
	
	Prd->ShareDisposition = 
			CmResourceShareShared;
			
	Prd->Flags = 0;
	
	Prd->u.Dma.Channel = 
				Device->DmaChannel;
	Prd++;

	//
	// Return address of next available PRD
	// in the FRD.
	//
	return Prd;
}

//++
// Function:
//		XxReleaseHardwareInfo
//
// Description:
//		This function deallocates space used to
//		hold hardware info gathered by the driver
//
// Arguments:
//		Pointer to linked-list of Config arrays
//
// Return Value:
//		(None)
//
//--
VOID
XxReleaseHardwareInfo(
	IN PCONFIG_ARRAY ConfigList
	)
{
	PCONFIG_ARRAY NextItem;

	while( ConfigList != NULL )
	{
		NextItem = ConfigList->NextConfigArray;
		ExFreePool( ConfigList );
		ConfigList = NextItem;
	}
}
