//++
// File Name:
//		transfer.c
//
// Contents:
//		Data transfer routines for 
//		packet-based slave DMA driver
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
static IO_ALLOCATION_ACTION
XxAdapterControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID MapRegisterBase,
    IN PVOID Context
    );

//++
// Function:
//		XxStartIo
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual data transfer.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		(None)
//--
VOID
XxStartIo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );
		
	PDEVICE_EXTENSION pDE =
		DeviceObject->DeviceExtension;

	PMDL Mdl = Irp->MdlAddress;
	ULONG MapRegsNeeded;
	NTSTATUS status;
		
	switch( IrpStack->MajorFunction ) {
		
		case IRP_MJ_WRITE:
		case IRP_MJ_READ:
			//
			// Set up bookkeeping values
			//
			pDE->BytesRequested = 
				MmGetMdlByteCount( Mdl );

			pDE->BytesRemaining = 
				pDE->BytesRequested;

			pDE->TransferVA = 
				MmGetMdlVirtualAddress( Mdl );

			//
			// Set the direction flag
			//
			if( IrpStack->MajorFunction 
				== IRP_MJ_WRITE )
			{
				pDE->WriteToDevice = TRUE;
			}
			else
			{	
				pDE->WriteToDevice = FALSE;
			}

			//
			// Calculate size of first partial
			// transfer
			//
			pDE->TransferSize = pDE->BytesRemaining;

			MapRegsNeeded = 
				ADDRESS_AND_SIZE_TO_SPAN_PAGES(
					pDE->TransferVA,
					pDE->TransferSize );

			if( MapRegsNeeded > 
				pDE->MapRegisterCount )
			{
				MapRegsNeeded = 
					pDE->MapRegisterCount;

				pDE->TransferSize = 
					MapRegsNeeded * PAGE_SIZE - 
						MmGetMdlByteOffset( Mdl );
			}

			//
			// Request Adapter object and
			// start a DMA operation
			//
			status = IoAllocateAdapterChannel(
						pDE->AdapterObject,
						DeviceObject,
						MapRegsNeeded,
						XxAdapterControl,
						pDE );

			//
			// If the call failed, it's because
			// there weren't enough mapping
			// registers. Fail the IRP and start
			// next one.
			//
			if( !NT_SUCCESS( status ))
			{
				Irp->IoStatus.Status = status;
				Irp->IoStatus.Information = 0;

				IoCompleteRequest(
					Irp,
					IO_NO_INCREMENT );

				IoStartNextPacket( 
					DeviceObject, 
					FALSE );
			}
			break;
		//
		// Should never get here -- just get rid
		// of the packet...
		//	
		default:
			Irp->IoStatus.Status =
						STATUS_NOT_SUPPORTED;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest(
				Irp,
				IO_NO_INCREMENT );
			IoStartNextPacket( DeviceObject, FALSE );
			break;
	}
}

//++
// Function:
//		XxAdapterControl
//
// Description:
//		This function sets up the first
//		part of a split transfer and
//		starts the DMA device.
//
// Arguments:
//		Device object
//		Current Irp
//		Map register base handle
//		Pointer to Device Extension
//
// Return Value:
//		KeepObject
//--
static IO_ALLOCATION_ACTION
XxAdapterControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID MapRegisterBase,
    IN PVOID Context
    )
{
	PDEVICE_EXTENSION pDE = Context;
	
	pDE->MapRegisterBase = MapRegisterBase;

	//
	// Flush CPU cache if necessary
	//
	KeFlushIoBuffers(
		Irp->MdlAddress,
		!pDE->WriteToDevice,
		TRUE );

	//
	// Set up the system DMA controller
	// attached to this device.
	//
	IoMapTransfer(
		pDE->AdapterObject,
		Irp->MdlAddress,
		pDE->MapRegisterBase,
		pDE->TransferVA,
		&pDE->TransferSize,
		pDE->WriteToDevice );

	//
	// Start the device
	//
	XxWriteControl( 
		pDE, 
		XX_CTL_INTENB | XX_CTL_DMA_GO );

	return KeepObject;
}

//++
// Function:
//		XxIsr
//
// Description:
//		This function processes interrupts from
//		the DMA device
//
// Arguments:
//		Pointer to the Interrupt object
//		Pointer to the Device Extension 
//
// Return Value:
//		TRUE if we handled the interrupt
//--
BOOLEAN
XxIsr(
	IN PKINTERRUPT Interrupt,
	IN PVOID ServiceContext
	)
{
	PDEVICE_EXTENSION pDE = ServiceContext;
	PDEVICE_OBJECT DeviceObject = pDE->DeviceObject;
	UCHAR Status = XxReadStatus( pDE );
	UCHAR Control;

	//
	// See if this device requested an interrupt
	//
	if(( Status & XX_STS_IRQ ) == 0 ) 
		return FALSE;
	
	//
	// It did. Stop the device and
	// disable its interrupts.
	//
	Control = XxReadControl( pDE );
	
	Control &= ~(XX_CTL_INTENB |
				  XX_CTL_DMA_GO );

	XxWriteControl(	pDE, Control);

	//
	// Save device status for DPC routine
	//
	pDE->DeviceStatus = Status;

	//
	// Do the rest of the work down
	// at DISPATCH_LEVEL IRQL
	//
	IoRequestDpc( 
		DeviceObject, 
		DeviceObject->CurrentIrp,
		(PVOID)pDE );

	return TRUE;
}

//++
// Function:
//		XxDpcForIsr
//
// Description:
//		This function performs sets up any
//		later transfers in a split DMA operation,
//		an does any low-IRQL post-processing 
//		of I/O requests
//
// Arguments:
//		Pointer to a DPC object
//		Pointer to the Device object
//		Pointer to the IRP for this request
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxDpcForIsr(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION pDE = Context;
	ULONG MapRegsNeeded;
	PMDL Mdl = Irp->MdlAddress;

	//
	// Flush data out of the Adapater
	// object cache.
	//
	IoFlushAdapterBuffers(
		pDE->AdapterObject,
		Mdl,
		pDE->MapRegisterBase,
		pDE->TransferVA,
		pDE->TransferSize,
		pDE->WriteToDevice );

	//
	// Check for device errors; if one occurred
	// complete the IRP and start the next one.
	// else continue...
	//
	if( !XX_STS_OK( pDE->DeviceStatus ))
	{
		//
		// Let someone else have the DMA channel
		//
		IoFreeAdapterChannel( pDE->AdapterObject );

		Irp->IoStatus.Status =	
						STATUS_DEVICE_DATA_ERROR;

		Irp->IoStatus.Information = 
					pDE->BytesRequested -
						pDE->BytesRemaining;
		//
		// Complete this request and
		// start the next
		//
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
 		IoStartNextPacket( DeviceObject, FALSE );
	}

	//
	// Previous transfer was successful. See if
	// we need to do another partial transfer.
	//
	pDE->BytesRemaining -= pDE->TransferSize;

	if( pDE->BytesRemaining > 0 )
	{
		//
		// Yes, update the pointer and try to
		// do all of it in one operation
		//
		pDE->TransferVA += pDE->TransferSize;
		pDE->TransferSize = pDE->BytesRemaining;

		MapRegsNeeded = 
			ADDRESS_AND_SIZE_TO_SPAN_PAGES(
					pDE->TransferVA,
					pDE->TransferSize );

		//
		// If the remainder of the buffer is more
		// than we can handle in one I/O. Reduce
		// our expectations.
		//
		if( MapRegsNeeded > pDE->MapRegisterCount )
		{
			MapRegsNeeded = pDE->MapRegisterCount;

			pDE->TransferSize = 
				MapRegsNeeded * PAGE_SIZE -
					BYTE_OFFSET( pDE->TransferVA );
		}

		//
		// Set up the system DMA controller
		// attached to this device.
		//
		IoMapTransfer(
			pDE->AdapterObject,
			Mdl,
			pDE->MapRegisterBase,
			pDE->TransferVA,
			&pDE->TransferSize,
			pDE->WriteToDevice );

		//
		// Start the device
		//
		XxWriteControl( 
			pDE, 
			XX_CTL_INTENB | XX_CTL_DMA_GO );
	}
	else
	{
		//
		// All done. Rlease the DMA channel.
		//
		IoFreeAdapterChannel( pDE->AdapterObject );

		//
		// Put final status in the IRP.
		//
		Irp->IoStatus.Status = STATUS_SUCCESS;

		//
		// Assume all of it got transferred since
		// there was no error.
		//
		Irp->IoStatus.Information = 
					pDE->BytesRequested;
		//
		// Complete this request and
		// start the next. (In a real driver,
		// you would pick an appropriate priority
		// boost.)
		//
		IoCompleteRequest( Irp, IO_DISK_INCREMENT );
 		IoStartNextPacket( DeviceObject, FALSE );
	}
}






	
	
	