//++
// File Name:
//		regcon.c
//
// Contents:
//		Routines that show how to retrieve 
//		user-supplied hardware information
//		from the Registry.
//--

//
// Driver-specific header files
//
#include "xxdriver.h"

//
// Forward declarations of local routines...
//
static NTSTATUS
XxFindNextDevice(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PUNICODE_STRING RegistryPath,
	IN PCONFIG_ARRAY ConfigArray
	);

static NTSTATUS
XxGetPortInfo(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	);

static NTSTATUS
XxGetInterruptInfo(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	);

static NTSTATUS
XxGetDmaInfo(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	);

static NTSTATUS
XxCheckForBus(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber
	);

static NTSTATUS 
XxBusCallback(
	IN PVOID Context,
	IN PUNICODE_STRING PathName,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PKEY_VALUE_FULL_INFORMATION *BusInfo,
	IN CONFIGURATION_TYPE CtrlrType,
	IN ULONG CtrlrNumber,
	IN PKEY_VALUE_FULL_INFORMATION *CtrlrInfo,
	IN CONFIGURATION_TYPE DeviceType,
	IN ULONG DeviceNumber,
	IN PKEY_VALUE_FULL_INFORMATION *DeviceInfo
	);


//
// If the compiler can handle it, save space
// by making various routines discardable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( init, XxGetHardwareInfo )
#pragma alloc_text( init, XxFindNextDevice )
#pragma alloc_text( init, XxGetPortInfo )
#pragma alloc_text( init, XxGetInterruptInfo )
#pragma alloc_text( init, XxGetDmaInfo )
#pragma alloc_text( init, XxCheckForBus )
#pragma alloc_text( init, XxBusCallback )
#endif

//++
// Function:
//		XxGetHardwareInfo
//
// Description:
//	This function extracts user-supplied information
//	about an ISA card from the Registry.
//
// Arguments:
//		Registry path of driver's service key
//		Pointer to variable that receives address of
//		  cofiguration list.
//
// Return Value:
//		STATUS_SUCCESS or some error condition
//--		
NTSTATUS
XxGetHardwareInfo(
	IN PUNICODE_STRING RegistryPath,
	OUT PCONFIG_ARRAY *ConfigList
	)
{
	NTSTATUS status;
	PCONFIG_ARRAY ConfigArray;
	INTERFACE_TYPE BusType;
	ULONG BusNumber;
	UNICODE_STRING TempString;

	//
	// Check for a bus we can use. Look for an ISA bus
	// first, then look for an EISA bus. If neither one
	// shows up, quit.
	//	
	BusType = Isa;
	BusNumber = 0;

	status = XxCheckForBus( Isa, BusNumber );

	if( !NT_SUCCESS( status ))
	{
		BusType = Eisa;
		status = XxCheckForBus( Eisa, BusNumber );
	}

	if( !NT_SUCCESS( status ))
	{
		*ConfigList = NULL;
		return STATUS_NO_SUCH_DEVICE;
	}

	//
	// We found a compatible bus. Allocate
	// space for the (single) Config array
	// that we'll be passing back to the
	// caller.
	//
	if(( ConfigArray = ExAllocatePool(
						PagedPool, 
						sizeof( CONFIG_ARRAY )))
		== NULL )
	{
		*ConfigList = NULL;
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	RtlZeroMemory( 
		ConfigArray, 
		sizeof( CONFIG_ARRAY ));

	*ConfigList = ConfigArray;
	ConfigArray->BusType = BusType;
	ConfigArray->BusNumber = BusNumber;

	//
	// Make a copy of the Registry path name 
	// and be sure it has a terminator at the 
	// end...
	//
	TempString.Length = 0;
	TempString.MaximumLength = 
				RegistryPath->Length + 
					sizeof( UNICODE_NULL );
				
	if(( TempString.Buffer = 
				ExAllocatePool(
					PagedPool,
					TempString.MaximumLength )) 
		== NULL )
	{
		*ConfigList = NULL;
		ExFreePool( ConfigArray );
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	RtlCopyUnicodeString( &TempString, RegistryPath );

	TempString.Buffer[ TempString.Length ] =
		UNICODE_NULL;
	
	//
	// Keep looping until we run out of device
	// slots or Registry entries, or until an
	// error occurs.
	//
	ConfigArray->Count = 0;

	while( ConfigArray->Count < XX_MAXIMUM_DEVICES )
	{
		status = XxFindNextDevice( 
					BusType,
					BusNumber,
					&TempString, 
					ConfigArray );

		if( !NT_SUCCESS( status )) break;

		ConfigArray->Count++;

	} // end while-loop

	ExFreePool( TempString.Buffer );

	if(	!NT_SUCCESS( status ) && 
		status != STATUS_OBJECT_NAME_NOT_FOUND )
	{
		*ConfigList = NULL;
		ExFreePool( ConfigArray );
		return status;
	}

	//
	// See if we found anything after all
	// that work
	//
	if( ConfigArray->Count == 0 )
	{
		*ConfigList = NULL;
		ExFreePool( ConfigArray );
		return STATUS_NO_SUCH_DEVICE;
	}

	//
	// Everything worked...
	//
	return STATUS_SUCCESS;
}

//++
// Function:
//		XxFindNextDevice
//
// Description:
//	This function extracts information about one
//	device from the driver's service key and stores
//	it in a slot in the Configuration block.
//
// Arguments:
//		Registry path of driver's service key
//		Configuration block
//
// Return Value:
//		STATUS_SUCCESS or some error condition
//--		
static NTSTATUS
XxFindNextDevice(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PUNICODE_STRING RegistryPath,
	IN PCONFIG_ARRAY ConfigArray
	)
{
	UNICODE_STRING SubPath;
	WCHAR PathNameBuffer[ 30 ];

	UNICODE_STRING	Number;
	WCHAR NumberBuffer[10];


	RTL_QUERY_REGISTRY_TABLE Table[6];
	NTSTATUS status;

	PDEVICE_BLOCK pDevice = 
		&ConfigArray->Device[ ConfigArray->Count ];

	//
	// Prepare to interrogate the Registry by
	// setting up the query-table
	//
	RtlZeroMemory( Table, sizeof(Table));

	//
	// Create a name string for the 
	// query table. Start by forming
	// the base path name
	//
	SubPath.Buffer = PathNameBuffer;
	SubPath.MaximumLength = sizeof( PathNameBuffer );
	SubPath.Length = 0;

   	RtlAppendUnicodeToString( 
		&SubPath,
   		L"Parameters\\Device" );

	//
	// Convert the device number into a string and
	// attach it to the end of the path name.
	//
	Number.Buffer = NumberBuffer;
	Number.MaximumLength = sizeof( NumberBuffer );
	Number.Length = 0;

	RtlIntegerToUnicodeString( 
			ConfigArray->Count, 
			10,				// base-10 conversion
			&Number ); 
	
	RtlAppendUnicodeStringToString( 
			&SubPath,
			&Number );

	//
	// Fabricate the query
	//
	Table[0].Name	= SubPath.Buffer;
	Table[0].Flags	= RTL_QUERY_REGISTRY_SUBKEY;

	Table[1].Name	= L"PORT";	// I/O port addr
	Table[1].Flags	= RTL_QUERY_REGISTRY_DIRECT;
	Table[1].EntryContext = 
				&pDevice->OriginalPortBase;

	Table[2].Name	= L"SPAN";	// Number of ports
	Table[2].Flags	= RTL_QUERY_REGISTRY_DIRECT;
	Table[2].EntryContext = 
				&pDevice->PortSpan;

	Table[3].Name	= L"IRQ";		// IRQ level
	Table[3].Flags	= RTL_QUERY_REGISTRY_DIRECT;
	Table[3].EntryContext = 
				&pDevice->OriginalIrql;

	Table[4].Name	= L"DMA";		// DMA channel
	Table[4].Flags	= RTL_QUERY_REGISTRY_DIRECT;
	Table[4].EntryContext = 
				&pDevice->DmaChannel;

	//
	// Query the Registry...
	//
	status = RtlQueryRegistryValues(
				RTL_REGISTRY_ABSOLUTE,
				RegistryPath->Buffer,
				Table,
				NULL, NULL );
	
	if( !NT_SUCCESS( status )) return status;

	//
	// Fix up and translate the information
	// from the Registry
	//
	status = XxGetPortInfo( 
				BusType, 
				BusNumber, 
				pDevice );

	if( !NT_SUCCESS( status )) return status;

	status = XxGetInterruptInfo( 
				BusType,
				BusNumber,
				pDevice );
	
	if( !NT_SUCCESS( status )) return status;

	status = XxGetDmaInfo( 
				BusType,
				BusNumber,
				pDevice );
	
	return status;
}

//++
// Function:
//		XxGetPortInfo
//
// Description:
//		This function fixes up I/O port infomation
//		pulled from the driver's Registry service key
//
// Arguments:
//		Bus type
//		Bus number
//		Pointer to this device's slot in Config array
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_XXX if error
//--
static NTSTATUS
XxGetPortInfo(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	)
{
	ULONG AddressSpace;
	PHYSICAL_ADDRESS TranslatedPortBase;
	//
	// Convert bus-relative port-information into NT 
	// system-mapped values, and save the results...
	//
	AddressSpace = 1; // Ports should be in I/O space.

	if( !HalTranslateBusAddress( 
			BusType,
			BusNumber,
			pDevice->OriginalPortBase,
			&AddressSpace,
			&TranslatedPortBase ))
	{
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	pDevice->PortBase =
				(PUCHAR)TranslatedPortBase.LowPart;

	return STATUS_SUCCESS;
}

//++
// Function:
//		XxGetInterruptInfo
//
// Description:
//		This function fixes up IRQ infomation
//		pulled from the driver's Registry service key
//
// Arguments:
//		Bus type
//		Bus number
//		Pointer to this device's slot in Config array
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_XXX if error
//--
static NTSTATUS
XxGetInterruptInfo(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	)
{
	//
	// Fill in the gaps by providing values for things 
	// that aren't in the Registry...
	//
	pDevice->InterruptMode = Latched;
	pDevice->OriginalVector = pDevice->OriginalIrql;
	pDevice->ShareVector = FALSE;
	pDevice->FloatingSave = FALSE;
	
	//
	// Convert bus-relative interrupt information into
	// NT system-mapped values, and save the results...
	//
	pDevice->SystemVector =
		HalGetInterruptVector(
			BusType,
			BusNumber,
		  	pDevice->OriginalIrql,
		  	pDevice->OriginalVector,
			&pDevice->Dirql,
			&pDevice->Affinity );	

	return STATUS_SUCCESS;
}

//++
// Function:
//		XxGetDmaInfo
//
// Description:
//		This function fixes up DMA infomation
//		pulled from the driver's Registry service key
//
// Arguments:
//		Bus type
//		Bus number
//		Pointer to this device's slot in Config array
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_XXX if error
//--
static NTSTATUS
XxGetDmaInfo(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PDEVICE_BLOCK pDevice
	)
{
	DEVICE_DESCRIPTION Descrip;

	RtlZeroMemory( 
		&Descrip, 
		sizeof( DEVICE_DESCRIPTION ));

	//
	// Create a description of the DMA hardware
	//
	Descrip.Version = DEVICE_DESCRIPTION_VERSION1;

	Descrip.Master				= FALSE;
	Descrip.ScatterGather		= FALSE;
	Descrip.DemandMode			= FALSE;
	Descrip.AutoInitialize		= FALSE;
	Descrip.Dma32BitAddresses	= FALSE;

	Descrip.InterfaceType	= BusType;
	Descrip.BusNumber		= BusNumber;

	Descrip.DmaChannel		= pDevice->DmaChannel;
	Descrip.MaximumLength	= XX_MAX_DMA_LENGTH;
	Descrip.DmaWidth		= Width16Bits;
	Descrip.DmaSpeed		= Compatible;

	//
	// Try to find the Adapter object
	//
	pDevice->MapRegisterCount = 
			(XX_MAX_DMA_LENGTH / PAGE_SIZE) + 2;

	pDevice->AdapterObject = 
					HalGetAdapter( 
						&Descrip, 
						&pDevice->MapRegisterCount );

	if( pDevice->AdapterObject == NULL )
		return STATUS_INSUFFICIENT_RESOURCES;
	else
		return STATUS_SUCCESS;
}

//++
// Function:
//		XxCheckForBus
//
// Description:
//		This function verifies the existence of a
//		particular bus-type and number.
//
// Arguments:
//		BusType -- Isa, Eisa, etc
//		BusNumber -- 0, 1, etc
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_XXX if error
//--
static NTSTATUS
XxCheckForBus(
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber
	)
{			
	return( IoQueryDeviceDescription(	
				&BusType, &BusNumber,
				NULL, NULL,
				NULL, NULL,
				XxBusCallback,
				NULL ));
}

//++
// Function:
//		XxBusCallback
//
// Description:
//		This is a dummy function. The fact that the
//		system calls it means that the bus type and
//		number both exist, so all that's necessary 
//		is to return STATUS_SUCCESS.
//
// Arguments:
//		(Unused)
//
// Return Value:
//		STATUS_SUCCESS always
//--
static NTSTATUS
XxBusCallback(
	IN PVOID Context,
	IN PUNICODE_STRING PathName,
	IN INTERFACE_TYPE BusType,
	IN ULONG BusNumber,
	IN PKEY_VALUE_FULL_INFORMATION *BusInfo,
	IN CONFIGURATION_TYPE CtrlrType,
	IN ULONG CtrlrNumber,
	IN PKEY_VALUE_FULL_INFORMATION *CtrlrInfo,
	IN CONFIGURATION_TYPE DeviceType,
	IN ULONG DeviceNumber,
	IN PKEY_VALUE_FULL_INFORMATION *DeviceInfo
	)
{
	return STATUS_SUCCESS;
}
