//++
// File Name:
//		thread.c
//
// Contents:
//		Driver thread routines 
//
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
// (None)

//++
// Function:
//		XxThreadMain
//
// Description:
//		This function runs in the context of a
//		driver-created system thread. It handles
//		all the processing of I/O requests.
//
// Arguments:
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxThreadMain(
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION DevExtension = Context;

	PDEVICE_OBJECT DeviceObject = 
			DevExtension->DeviceObject;

	PLIST_ENTRY ListEntry;
	PIRP Irp;
	CCHAR PriorityBoost;

	KeSetPriorityThread(
		KeGetCurrentThread(),
		LOW_REALTIME_PRIORITY );

	//
	// Now enter the main IRP-processing loop
	//
	while( TRUE )
	{
		//
		// Wait indefinitely for an IRP to
		// appear in the work queue or for
		// the Unload routine to stop the
		// thread.
		//
		KeWaitForSingleObject(
			&DevExtension->IrpQueueSemaphore,
			Executive,
			KernelMode,
			FALSE,
			NULL );

		//
		// See if thread was awakened because
		// driver is unloading itself...
		//
		if( DevExtension->ThreadShouldStop )
			PsTerminateSystemThread( STATUS_SUCCESS );

		//
		// It must be a real request. Get an IRP
		//
		ListEntry = 
			ExInterlockedRemoveHeadList(
				&DevExtension->IrpQueueListHead,
				&DevExtension->IrpQueueSpinLock );

		Irp = CONTAINING_RECORD(
				ListEntry,
				IRP,
				Tail.Overlay.ListEntry );

		//
		// Process the IRP. This is a synchronous
		// operation, so this function doesn't return
		// until it's time to get rid of the IRP.
		//
		PriorityBoost = 
			XxPerformDataTransfer(
				DeviceObject, 
				Irp );

		//
		// Release the IRP and go back to the
		// top of the loop to see if there's
		// another request waiting.
		//
		IoCompleteRequest( Irp, PriorityBoost );

	} // end of while-loop
}

//++
// Function:
//		XxKillThread
//
// Description:
//		This function tells a driver thread that
//		it's time to go away. It doesn't return
//		to the caller until the Thread object
//		has gone into the Signaled state.
//
// Arguments:
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxKillThread(
	IN PDEVICE_EXTENSION pDE
	)
{
	//
	// Set the Stop flag
	//
	pDE->ThreadShouldStop = TRUE;

	//
	// Make sure the thread wakes up 
	//
	KeReleaseSemaphore(
		&pDE->IrpQueueSemaphore,
		0,			// No priority boost
		1,			// Increment semaphore by 1
		TRUE );		// WaitForXxx after this call

	//
	// Wait for the thread to terminate
	//
	KeWaitForSingleObject(
		&pDE->ThreadObject,
		Executive,
		KernelMode,
		FALSE,
		NULL );

	ObDereferenceObject( 
		&pDE->ThreadObject );
}

