//++
// File Name:
//		transfer.c
//
// Contents:
//		Data transfer routines for 
//		packet-based slave DMA driver
//		that uses a system thread.
//
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
static NTSTATUS
XxAcquireAdapterObject(
	IN PDEVICE_EXTENSION pDE,
	IN ULONG MapRegsNeeded
	);

static IO_ALLOCATION_ACTION
XxAdapterControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID MapRegisterBase,
    IN PVOID Context
    );

static NTSTATUS
XxPerformSynchronousTransfer(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

//++
// Function:
//		XxProcessAnIrp
//
// Description:
//		This function moves an entire buffer of
//		data to or from the device. This may
//		include splitting the transfer over several
//		device operations if there aren't enough
//		mapping registers to handle it all at once.
//		It doesn't return to the caller until
//		everything is done.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		Boost value to be applied to this IRP
//		when it's completed. It will be one of
//		the following:
//			IO_NO_INCREMENT
//			IO_DISK_INCREMENT
//
//--
CCHAR
XxPerformDataTransfer(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );
		
	PDEVICE_EXTENSION pDE =
		DeviceObject->DeviceExtension;

	PMDL Mdl = Irp->MdlAddress;
	ULONG MapRegsNeeded;
	NTSTATUS status;

	//
	// Set the I/O direction flag
	//
	if( IrpStack->MajorFunction 
		== IRP_MJ_WRITE )
	{
		pDE->WriteToDevice = TRUE;
	}
	else
	{	
		pDE->WriteToDevice = FALSE;
	}

	//
	// Set up bookkeeping values
	//
	pDE->BytesRequested = 
			MmGetMdlByteCount( Mdl );

	pDE->BytesRemaining = 
			pDE->BytesRequested;

	pDE->TransferVA = 
			MmGetMdlVirtualAddress( Mdl );

	//
	// Flush CPU cache if necessary
	//
	KeFlushIoBuffers(
		Irp->MdlAddress,
		!pDE->WriteToDevice,
		TRUE );

	//
	// Calculate size of first partial
	// transfer
	//
	pDE->TransferSize = pDE->BytesRemaining;

	MapRegsNeeded = 
		ADDRESS_AND_SIZE_TO_SPAN_PAGES(
			pDE->TransferVA,
			pDE->TransferSize );

	if( MapRegsNeeded > pDE->MapRegisterCount )
	{
		MapRegsNeeded = 
			pDE->MapRegisterCount;

		pDE->TransferSize = 
			MapRegsNeeded * PAGE_SIZE - 
			MmGetMdlByteOffset( Mdl );
	}

	//
	// Aacquire the adapter object.
	//
	status = XxAcquireAdapterObject( 
				pDE, 
				MapRegsNeeded );
	if( !NT_SUCCESS( status ))
	{
		Irp->IoStatus.Status = status;
		Irp->IoStatus.Information = 0;
		return IO_NO_INCREMENT;
	}

	//
	// Try to perform the first partial
	// transfer
	//
	status = 
		XxPerformSynchronousTransfer( 
			DeviceObject,
			Irp );

	if( !NT_SUCCESS( status ))
	{
		IoFreeAdapterChannel( pDE->AdapterObject );
		Irp->IoStatus.Status = status;
		Irp->IoStatus.Information = 0;
		return IO_NO_INCREMENT;
	}

	//
	// It worked. Update the bookkeeping
	// information.
	//
	pDE->TransferVA += pDE->TransferSize;
	pDE->BytesRemaining -= pDE->TransferSize;

	//
	// Loop through all the partial transfer 
	// operations for this request.
	//
	while( pDE->BytesRemaining >0 )
	{
		//
		// Try to do all of it in one operation
		//
		pDE->TransferSize = pDE->BytesRemaining;

		MapRegsNeeded = 
			ADDRESS_AND_SIZE_TO_SPAN_PAGES(
					pDE->TransferVA,
					pDE->TransferSize );

		//
		// If the remainder of the buffer is more
		// than we can handle in one I/O. Reduce
		// our expectations.
		//
		if( MapRegsNeeded > pDE->MapRegisterCount )
		{
			MapRegsNeeded = pDE->MapRegisterCount;

			pDE->TransferSize = 
				MapRegsNeeded * PAGE_SIZE -
					BYTE_OFFSET( pDE->TransferVA );
		}

		//
		// Try to perform a device operation.
		//
		status = 
			XxPerformSynchronousTransfer( 
				DeviceObject,
				Irp );

		if( !NT_SUCCESS( status )) break;

		//
		// It worked. Update the bookkeeping
		// information for the next cycle.
		//
		pDE->TransferVA += pDE->TransferSize;
		pDE->BytesRemaining -= pDE->TransferSize;
	}

	//
	// After the last partial transfer is done,
	// release the DMA Adapter object .
	//
	IoFreeAdapterChannel( pDE->AdapterObject );

	//
	// Send the IRP back to the caller. Its final
	// status is the status of the last transfer
	// operation. 
	//
	Irp->IoStatus.Status = status;
	Irp->IoStatus.Information = 
				pDE->BytesRequested -
					pDE->BytesRemaining;
	//
	// Since there has been at least one I/O
	// operation, give the IRP a priority boost.
	//
	return IO_DISK_INCREMENT;
}

//++
// Function:
//		XxAcquireAdapterObject
//
// Description:
//		This function grabs the Adapter object
//		associated with our device. It doesn't
//		return until either we own the Adapter
//		object or there's been an error.
//
// Arguments:
//		Pointer to Device Extension
//		Number of mapping registers needed
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_INSUFFICIENT_RESOURCES
//
//--
static NTSTATUS
XxAcquireAdapterObject(
	IN PDEVICE_EXTENSION pDE,
	IN ULONG MapRegsNeeded
	)
{
	KIRQL OldIrql;
	NTSTATUS status;

	//
	// We must be at DISPATCH_LEVEL in order
	// to request the Adapter object
	//
	KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );

	status = IoAllocateAdapterChannel(
				pDE->AdapterObject,
				pDE->DeviceObject,
				MapRegsNeeded,
				XxAdapterControl,
				pDE );

	KeLowerIrql( OldIrql );

	//
	// If the call failed, it's because there
	// weren't enough mapping registers.
	//
	if( !NT_SUCCESS( status ))
	{
		return status;
	}
			
	//
	// Stop and wait for the Adapter Control
	// routine to set the Event object. This is
	// our signal that the Adapter object is ours.
	//
	KeWaitForSingleObject(
		&pDE->AdapterObjectIsAcquired,
		Executive,
		KernelMode,
		FALSE,
		NULL );

	return STATUS_SUCCESS;
}

//++
// Function:
//		XxAdapterControl
//
// Description:
//		The I/O Manager calls this function
//		when the DMA hardware becomes available.
//		This routine then signals the thread that
//		it's alright to start a data transfer.
//
// Arguments:
//		Device object
//		Current Irp
//		Map register base handle
//		Pointer to Device Extension
//
// Return Value:
//		KeepObject
//
//--
static IO_ALLOCATION_ACTION
XxAdapterControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID MapRegisterBase,
    IN PVOID Context
    )
{
	PDEVICE_EXTENSION pDE = Context;
	
	//
	// Save the handle to the mapping
	// registers. The thread will need it
	// to set up data transfers.
	//
	pDE->MapRegisterBase = MapRegisterBase;

	//
	// Let the thread know that its Device
	// object the Adapter object
	//
	KeSetEvent( 
		&pDE->AdapterObjectIsAcquired,
		0,
		FALSE );

	return KeepObject;
}


//++
// Function:
//		XxPerformSynchronousTransfer
//
// Description:
//		This function performs a single data
//		transfer operation. It doesn't return
//		to the caller until the transfer is
//		finished.
//
// Arguments:
//		Pointer to Device Object
//		Pointer to the IRP
//
// Return Value:
//		STATUS_SUCCESS
//		STATUS_DEVICE_DATA_ERROR
//
//--
static NTSTATUS
XxPerformSynchronousTransfer(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PDEVICE_EXTENSION pDE = 
		DeviceObject->DeviceExtension;

	//
	// Set up the system DMA controller
	// attached to this device.
	//
	IoMapTransfer(
		pDE->AdapterObject,
		Irp->MdlAddress,
		pDE->MapRegisterBase,
		pDE->TransferVA,
		&pDE->TransferSize,
		pDE->WriteToDevice );

	//
	// Start the device
	//
	XxWriteControl( 
		pDE, 
		XX_CTL_INTENB | XX_CTL_DMA_GO );

	//
	// The DPC routine will set an Event
	// object when the I/O operation is
	// done. Stop here and wait for it.
	//
	KeWaitForSingleObject(
		&pDE->DeviceOperationComplete,
		Executive,
		KernelMode,
		FALSE,
		NULL );

	//
	// Flush data out of the Adapater
	// object cache.
	//
	IoFlushAdapterBuffers(
		pDE->AdapterObject,
		Irp->MdlAddress,
		pDE->MapRegisterBase,
		pDE->TransferVA,
		pDE->TransferSize,
		pDE->WriteToDevice );

	//
	// Check for device errors
	//
	if( !XX_STS_OK( pDE->DeviceStatus ))
		return STATUS_DEVICE_DATA_ERROR;
	else
		return STATUS_SUCCESS;
}

//++
// Function:
//		XxIsr
//
// Description:
//		This function processes interrupts from
//		the DMA device
//
// Arguments:
//		Pointer to the Interrupt object
//		Pointer to the Device Extension 
//
// Return Value:
//		TRUE if we handled the interrupt
//--
BOOLEAN
XxIsr(
	IN PKINTERRUPT Interrupt,
	IN PVOID ServiceContext
	)
{
	PDEVICE_EXTENSION pDE = ServiceContext;
	PDEVICE_OBJECT DeviceObject = pDE->DeviceObject;
	UCHAR Status = XxReadStatus( pDE );
	UCHAR Control;

	//
	// See if this device requested an interrupt
	//
	if(( Status & XX_STS_IRQ ) == 0 ) 
		return FALSE;
	
	//
	// It did. Stop the device and
	// disable its interrupts.
	//
	Control = XxReadControl( pDE );
	
	Control &= ~(XX_CTL_INTENB |
				  XX_CTL_DMA_GO );

	XxWriteControl(	pDE, Control);

	//
	// Save device status for DPC routine
	//
	pDE->DeviceStatus = Status;

	//
	// Do the rest of the work down
	// at DISPATCH_LEVEL IRQL
	//
	IoRequestDpc( 
		DeviceObject, 
		NULL,
		(PVOID)pDE );

	return TRUE;
}

//++
// Function:
//		XxDpcForIsr
//
// Description:
//		This DPC routine sets an Event object
//		into the Signaled state. This wakes up
//		the thread waiting for the I/O operation
//		to complete.
//
// Arguments:
//		Pointer to a DPC object
//		Pointer to the Device object
//		Pointer to the IRP for this request
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxDpcForIsr(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION pDE = Context;

	KeSetEvent( 
		&pDE->DeviceOperationComplete,
		0,
		FALSE );

	return;
}






	
	
	