// File Name:
//		xxdriver.h
//
// Contents:
//		Constants, structures, and function
//		declarations specific to this driver.
//
#ifndef __XX_DRIVER__
#define __XX_DRIVER__

//
// Header files
//
#include <ntddk.h>
#include "hardware.h"

//
// Various constants
//
#define	XX_NT_DEVICE_NAME		L"\\Device\\xx"
#define	XX_WIN32_DEVICE_NAME	L"\\DosDevices\\XX"
#define	XX_DRIVER_NAME			L"XXDRIVER"

#define	XX_MAXIMUM_DEVICES	10
#define	XX_MAX_NAME_LENGTH	50

//++
// Description:
//		Driver-defined structure used to hold 
//		miscellaneous device information.
//
// Access:
//		Allocated from NON-PAGED POOL
//		Available at any IRQL
//--
typedef struct _DEVICE_EXTENSION 
{
	PDEVICE_OBJECT DeviceObject; // Back pointer
	
	ULONG NtDeviceNumber;	// Zero-based device num	

	PUCHAR PortBase;		// First control register
	
	PKINTERRUPT pInterrupt;	// Interrupt object
	
	//
	// Items relating to DMA hardware
	//
	PADAPTER_OBJECT AdapterObject;
	ULONG MapRegisterCount;

	//
	// Bookkeeping items used during
	// DMA transfers
	//
	PVOID MapRegisterBase;

	ULONG BytesRequested;
	ULONG BytesRemaining;

	ULONG TransferSize;
	PUCHAR TransferVA;

	BOOLEAN WriteToDevice;
	UCHAR DeviceStatus;

	//
	// Items relating to the Thread object
	//
	PETHREAD ThreadObject;
	BOOLEAN ThreadShouldStop;

	//
	// Events that signal interesting
	// occurences.
	//
	KEVENT AdapterObjectIsAcquired;
	KEVENT DeviceOperationComplete;

	//
	// Work queue of requests
	// to be processed
	//
	KSEMAPHORE IrpQueueSemaphore;
	LIST_ENTRY IrpQueueListHead;
	KSPIN_LOCK IrpQueueSpinLock;

} DEVICE_EXTENSION, *PDEVICE_EXTENSION;


//++
// Description:
//		Temporary struct used during driver 
//		initialization to hold config data about
//		a single device.
//
// Access:
//		Allocated from PAGED POOL
//		Available ONLY at IRQL PASSIVE_LEVEL
//--
typedef struct _DEVICE_BLOCK
{
	//
	// Original values pulled from the Registry
	//
	PHYSICAL_ADDRESS OriginalPortBase;
	ULONG PortSpan;	

	ULONG OriginalIrql;
	ULONG OriginalVector;
	KINTERRUPT_MODE InterruptMode;
	BOOLEAN ShareVector;
	BOOLEAN FloatingSave;
	
	ULONG DmaChannel;

	//
	// Converted values that will be used by
	// the driver		
	//
	PUCHAR PortBase;	// First control register

	ULONG SystemVector;	// Interrupt items
	KIRQL Dirql;
	KAFFINITY Affinity;

	PADAPTER_OBJECT AdapterObject;	// DMA items
	ULONG MapRegisterCount;
} DEVICE_BLOCK, *PDEVICE_BLOCK;

//++
// Description:
//		Temporary struct used during driver 
//		initialization to hold hardware 
//		configuration data about all devices.
//
// Access:
//		Allocated from PAGED POOL
//		Available ONLY at IRQL PASSIVE_LEVEL
//--
typedef struct _CONFIG_ARRAY
{
	//
	// We keep a list of these arrays, one
	// for each bus-type/bus-number combination
	// where we find our hardware.
	//
	struct _CONFIG_ARRAY *NextConfigArray;

	//
	// The bus to which all the devices are
	// attached.
	//
	INTERFACE_TYPE BusType;
	ULONG BusNumber;

	//
	// Number of devices in the block
	//
	ULONG	Count;
	
	//
	// One array-element for each device
	//
	DEVICE_BLOCK Device[ XX_MAXIMUM_DEVICES ];
	
} CONFIG_ARRAY, *PCONFIG_ARRAY;


//
// Prototypes for globally defined functions...
//
// init.c
//
NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	);

//
// unload.c
//
VOID
XxDriverUnload(
	IN PDRIVER_OBJECT DriverObject
	);

VOID
XxReleaseHardware(
	IN PDRIVER_OBJECT DriverObject
	);

//
// regcon.c
//
NTSTATUS
XxGetHardwareInfo(
	IN PUNICODE_STRING RegistryPath,
	OUT PCONFIG_ARRAY *ConfigList
	);

//
// resalloc.c
//
NTSTATUS
XxReportHardwareUsage(
	IN PDRIVER_OBJECT DriverObject,
	IN PCONFIG_ARRAY ConfigList
	);

VOID
XxReleaseHardwareInfo(
	IN PCONFIG_ARRAY ConfigList
	);

//
// dispatch.c
//
NTSTATUS
XxDispatchOpenClose(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

NTSTATUS
XxDispatchReadWrite(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

//
// transfer.c
//
CCHAR
XxPerformDataTransfer(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	);

BOOLEAN
XxIsr(
	IN PKINTERRUPT Interrupt,
	IN PVOID ServiceContext
	);

VOID
XxDpcForIsr(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	);

//
// thread.c
//
VOID
XxThreadMain(
	IN PVOID Context
	);

VOID
XxKillThread(
	IN PDEVICE_EXTENSION pDE
	);

//
// hardware.c
//
BOOLEAN
XxInitializeHardware(
	PDEVICE_EXTENSION pDE
	);

BOOLEAN
XxEnableInterrupts(
	PDEVICE_EXTENSION pDE
	);

BOOLEAN
XxDisableInterrupts(
	PDEVICE_EXTENSION pDE
	);

#endif	// __XX_DRIVER__

