//++
// File Name:
//		complete.c
//
// Contents:
//		I/O Completion routines
//
//--

//
// Driver-specific header files...
//
#include "yydriver.h"

//
// Forward declarations of local functions
//
// (None)

//
// If the compiler can handle it, save space
// by making various routines pageable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, YyGenericCompletion )
#pragma alloc_text( page, YyWriteCompletion )
#endif // ALLOC_PRAGMA

//++
// Function:
//		YyGenericCompletion
//
// Description:
//		This is the generic I/O Completion routine
//		used by most IRP_MJ_XXX functions.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to IRP being completed
//		Context given to IoSetCompletionRoutine
//
// Return Value:
//		STATUS_SUCCESS
//
//--
NTSTATUS
YyGenericCompletion(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	KdPrint(( "[YyGenericCompletion] IRP: %x\n",
		Irp ));

	if( Irp->PendingReturned )
	{
		IoMarkIrpPending(Irp);
	}
	return STATUS_SUCCESS;
}

//++
// Function:
//		YyWriteCompletion
//
// Description:
//		This is the I/O Completion routine
//		used for IRP_MJ_WRITE operations.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to IRP being completed
//		Context given to IoSetCompletionRoutine
//
// Return Value:
//		STATUS_MORE_PROCESSING_REQUIRED
//		STATUS_SUCCESS
//
//--
NTSTATUS
YyWriteCompletion(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION FilterExtension =
		DeviceObject->DeviceExtension;

	PIO_STACK_LOCATION IrpStack =
		IoGetCurrentIrpStackLocation( Irp );

	PIO_STACK_LOCATION NextIrpStack =
		IoGetNextIrpStackLocation( Irp );

	ULONG TransferSize = 
		Irp->IoStatus.Information;

	ULONG BytesRequested = 
		IrpStack->Parameters.Write.Length;

	ULONG BytesRemaining = 
		(ULONG)IrpStack->
			Parameters.Write.ByteOffset.HighPart;

	ULONG MaxTransfer =
		FilterExtension->BufferInfo.MaxWriteLength;

	NTSTATUS status;

	KdPrint((
		"[YyWriteCompletion] IRP: %x "
		"(%d bytes transferred.)\n",
		Irp, 
		TransferSize ));

	//
	// If this transfer worked, reduce the count of
	// bytes left to send. Save the new count in an
	// unused part of the filter driver's I/O stack
	// location.
	//
	if( NT_SUCCESS( Irp->IoStatus.Status ))
	{
		BytesRemaining -= TransferSize;

		IrpStack->
			Parameters.Write.
				ByteOffset.HighPart = 
							BytesRemaining;
	}

	//
	// If everything worked and there's more
	// data to transfer, set up the next partial
	// transfer operation.
	//
	if( NT_SUCCESS( Irp->IoStatus.Status )
		&& BytesRemaining > 0 )
	{
		//
		// Update the pointer into the system buffer
		// to account for the data transfer that just
		// completed.
		//
		(PUCHAR)Irp->
			AssociatedIrp.SystemBuffer += 
									TransferSize;

		//
		// Calculate the size of the next partial
		// transfer. Start by assuming it can all
		// be done in a single operation. Reduce
		// that expectation if it proves to be too
		// optimistic.
		//
		TransferSize = BytesRemaining;

		if( TransferSize > MaxTransfer )
		{
			TransferSize = MaxTransfer;
		}

		//
		// Set up I/O stack location for
		// target device.
		//
		NextIrpStack->MajorFunction = IRP_MJ_WRITE;

		NextIrpStack->
			Parameters.Write.Length = 
								TransferSize;

		KdPrint(( 
			"\tRecycling IRP: %8x (%d bytes).\n",
			Irp, 
			TransferSize ));

		//
		// Set up Completion routine to catch the
		// operation when it finishes.
		//
		IoSetCompletionRoutine(
			Irp, 
			YyWriteCompletion, 
			NULL, 
			TRUE, TRUE, TRUE );

		//
		// Pass the IRP to the target. Return value
		// doesn't matter because the I/O Completion
		// routine will clean up after everything.
		//
		IoCallDriver( 
				FilterExtension->TargetDevice,
				Irp );

		//
		// Prevent any further processing of this IRP
		// by the IoCompleteRequest mechanism. After
		// all, the IRP has just been given away...
		//
		return STATUS_MORE_PROCESSING_REQUIRED;
	}
	//
	// There was either an error or all the
	// bytes are transferred. Either way, get
	// rid of the IRP.
	//
	else
	{
		//
		// Restore the original buffer pointer
		//
		Irp->AssociatedIrp.SystemBuffer = 
			(PVOID)IrpStack->
					Parameters.Write.
						ByteOffset.LowPart;

		//
		// Calculate true number of bytes 
		// transfered in all the partial
		// operations completed so far.
		//
		Irp->IoStatus.Information = 
			BytesRequested - BytesRemaining;

		//
		// See if I/O stack location should be
		// marked pending, then send the IRP on
		// its way to any higher-level drivers.
		//
		if( Irp->PendingReturned )
		{
			IoMarkIrpPending(Irp);
		}
		return STATUS_SUCCESS;
	}
}
