//++
// File Name:
//		dispatch.c
//
// Contents:
//		Dispatch routines for driver
//--

//
// Driver-specific header files
//
#include "yydriver.h"

//
// Forward declarations of local routines
//
// (None)

//
// If the compiler can handle it, save space
// by making various routines pageable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, YyDispatchWrite )
#pragma alloc_text( page, YyDispatchDeviceIoControl )
#pragma alloc_text( page, YyDispatchPassThrough )
#endif //ALLOC_PRAGMA

//++
// Function:
//		YyDispatchWrite
//
// Description:
//		This function dispatches WriteFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
YyDispatchWrite(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PDEVICE_EXTENSION FilterExtension = 
			DeviceObject->DeviceExtension;

	PIO_STACK_LOCATION IrpStack = 
			IoGetCurrentIrpStackLocation( Irp );

	PIO_STACK_LOCATION NextIrpStack =
			IoGetNextIrpStackLocation( Irp );

	ULONG MaxTransfer = 
			FilterExtension->
					BufferInfo.MaxWriteLength;

	ULONG BytesRequested =
			IrpStack->Parameters.Write.Length;

	KdPrint(( 
		"\n[YyDispatchWrite] "
		"IRP: %8x (%d bytes).\n",
		Irp, BytesRequested ));

	//
	// Check for zero-length transfers
	//
	if( BytesRequested == 0 )
	{
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}

	//
	// If the length is within the lower driver's
	// acceptable limits, just send the request on.
	//
	if( BytesRequested <= MaxTransfer )
	{
		return YyDispatchPassThrough( 
							DeviceObject, 
							Irp );
	}

	//
	// Otherwise, set up the lower driver's I/O stack
	// location to transfer as much as possible in a
	// single operation.
	//
	NextIrpStack->
		MajorFunction = IRP_MJ_WRITE;

	NextIrpStack->
		Parameters.Write.Length = MaxTransfer;

	//
	// Use the high part of the ByteOffset field in
	// the filter driver's I/O stack location to hold
	// the number of bytes remaining in the original 
	// caller's request. Initially, this is the same
	// as the number of bytes requested in the whole
	// transfer.
	//
	IrpStack->
		Parameters.Write.
			ByteOffset.HighPart = BytesRequested;
	//
	// Save the original system buffer address in
	// the low (unsigned) 32-bits of the ByteOffset
	// field of the filter driver's I/O stack location.
	// This is okay because it isn't being used for
	// anything else.
	//
	IrpStack->
		Parameters.Write.
			ByteOffset.LowPart = (ULONG)Irp->
									AssociatedIrp.
										SystemBuffer;

	//
	// Set up Completion routine to continue
	// working on the request. All the necessary
	// context is stored in the IRP somewhere, so
	// there's no need to keep anything else.
	//
	IoSetCompletionRoutine(
		Irp, 
		YyWriteCompletion, 
		NULL, 
		TRUE, TRUE, TRUE );

	//
	// Pass the IRP to the target device
	//
	return IoCallDriver( 
				FilterExtension->TargetDevice,
				Irp );
}

//++
// Function:
//		YyDispatchDeviceIoControl
//
// Description:
//		This function dispatches DeviceIoControl
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
YyDispatchDeviceIoControl(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION IrpStack =
		IoGetCurrentIrpStackLocation( Irp );

	PXX_BUFFER_SIZE_INFO BufferInfo;

	//
	// Intercept the buffer-size IOCTL code and
	// tell the caller that there's no limit.
	//
	if( IrpStack->
			Parameters.
				DeviceIoControl.IoControlCode
		== IOCTL_XX_GET_MAX_BUFFER_SIZE )
	{
		BufferInfo = 
			(PXX_BUFFER_SIZE_INFO)Irp->
					AssociatedIrp.SystemBuffer;

		BufferInfo->
			MaxWriteLength = XX_NO_BUFFER_LIMIT;

		BufferInfo->
			MaxReadLength = XX_NO_BUFFER_LIMIT;

		Irp->IoStatus.Information = 
					sizeof( XX_BUFFER_SIZE_INFO );

		Irp->IoStatus.Status = STATUS_SUCCESS;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	//
	// If it's any other kind of IOCTL, just send
	// it on to the lower driver.
	//
	else
	{
		return YyDispatchPassThrough( 
							DeviceObject,
							Irp );
	}
}

//++
// Function:
//		YyDispatchPassThrough
//
// Description:
//		This function simply passes a request
//		on to the lower driver.
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
YyDispatchPassThrough(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PDEVICE_EXTENSION FilterExtension = 
			DeviceObject->DeviceExtension;

	PIO_STACK_LOCATION IrpStack = 
			IoGetCurrentIrpStackLocation( Irp );

	PIO_STACK_LOCATION NextIrpStack =
			IoGetNextIrpStackLocation( Irp );

	NTSTATUS status;

	KdPrint(( 
		"\n[YyDispatchPassThrough] "
		"IRP: %8x; "
		"MajorFunction: %d\n",
		Irp, 
		IrpStack->MajorFunction ));

	//
	// Copy args to next level
	//
	*NextIrpStack = *IrpStack;

	//
	// Set up Completion routine to handle
	// marking the IRP pending.
	//
	IoSetCompletionRoutine(
		Irp, 
		YyGenericCompletion, 
		NULL, 
		TRUE, TRUE, TRUE );

	//
	// Pass the IRP to the target
	//
	return IoCallDriver( 
				FilterExtension->TargetDevice,
				Irp );
}

