//++
// File Name:
//		init.c
//
// Contents:
//		Driver initialization code
//--

//
// Driver-specific header files...
//
#include "yydriver.h"

//
// Forward declarations of local functions
//
static NTSTATUS
YyAttachFilter(
	IN PDRIVER_OBJECT FilterDriver,
	IN PDEVICE_OBJECT TargetDevice,
	IN PXX_BUFFER_SIZE_INFO BufferInfo
	);

static VOID
YyGetBufferLimits(
	IN PDEVICE_OBJECT TargetDevice,
	IN OUT PXX_BUFFER_SIZE_INFO BufferInfo
	);

//
// If the compiler can handle it, make the DriverEntry
// routine discardable, so that it doesn't waste space
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( init, DriverEntry )
#pragma alloc_text( init, YyAttachFilter )
#pragma alloc_text( init, YyGetBufferLimits )
#endif // ALLOC_PRAGMA

//++
// Function:
//		DriverEntry
//
// Description:
//		This function initializes the driver, and
//		creates various NT objects needed to process
//		I/O requests.
//
// Arguments:
//		Pointer to the Driver object
//		Registry path string for driver service key
//
// Return Value:
//		NTSTATUS signaling success or failure
//--
NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	)
{

	PDEVICE_OBJECT TargetDevice;
	UNICODE_STRING TargetDeviceName;
	PDRIVER_OBJECT TargetDriver;
	PDRIVER_DISPATCH EmptyDispatchValue;

	XX_BUFFER_SIZE_INFO BufferInfo;
	PFILE_OBJECT FileObject;

	NTSTATUS status;
	ULONG i;

	//
	// Get the contents of an empty slot in the
	// MajorFunction table. (This is an address in the
	// I/O Manager rather than just a NULL, so we have
	// to find it on the local machine.)
	// (The function is called _IopInvalidDeviceRequest)
	// We can use any slot that we haven't filled
	// in to get the value that goes in an unused
	// slot.
	//
	EmptyDispatchValue = 
		DriverObject->MajorFunction[ IRP_MJ_CREATE ];

	//
	// Export other driver entry points...
	//
 	DriverObject->DriverUnload = YyDriverUnload;

	//
	// Fill in the Dispatch slots intercepted
	// by the filter driver.
	//
	DriverObject->
		MajorFunction[ IRP_MJ_WRITE ] =
			YyDispatchWrite;

	DriverObject->
		MajorFunction[ IRP_MJ_DEVICE_CONTROL ] =
			YyDispatchDeviceIoControl;

     //
    // Create counted string version of 
	// target device name.
    //
    RtlInitUnicodeString(
		&TargetDeviceName,
		TARGET_DEVICE_NAME );

	//
	// Get a pointer to its Device object
	//
	status = IoGetDeviceObjectPointer(
				&TargetDeviceName,
				FILE_ALL_ACCESS,
				&FileObject,
				&TargetDevice );

	if( !NT_SUCCESS( status ))
	{
		return status;
    }

	//
	// Find out the biggest buffers the
	// target device can handle.
	//
	YyGetBufferLimits( TargetDevice, &BufferInfo );

	TargetDriver = TargetDevice->DriverObject;
	

	//
	// See which IRP_MJ_XXX function codes the target
	// device responds to. If the filter driver doesn't
	// intercept a given IRP_MJ_XXX code, fill the
	// corresponding slot in the filter's MajorFunction
	// table with the address of the pass-thru Dispatch
	// routine.
	//
	for( i=0; i<=IRP_MJ_MAXIMUM_FUNCTION; i++ )
	{
		if(( TargetDriver->MajorFunction[i]
			!= EmptyDispatchValue )
		&& ( DriverObject->MajorFunction[i]
			== EmptyDispatchValue ))
		{
			DriverObject->MajorFunction[i] =
							YyDispatchPassThrough;
		}
	}

	//
	// Run the list of Device objects attached to
	// the target driver. For each one, attach a
	// filter Device.
	//
	TargetDevice = TargetDriver->DeviceObject;

	while( TargetDevice != NULL )
	{
		status = YyAttachFilter(
					DriverObject, 
					TargetDevice, 
					&BufferInfo );

		if( !NT_SUCCESS( status ))
		{
			YyDriverUnload( DriverObject );
			break;
		}
		TargetDevice = TargetDevice->NextDevice;
	}

	//
	// Decrement reference count on unused
	// File object
	//
	ObDereferenceObject( FileObject );

	//
	// Return most recent status value. This is
	// either STATUS_SUCCESS or some error code
	// returned by YyAttachFilter.
	//
	return status;
}

//++
// Function:
//		YyAttachFilter
//
// Description:
//		This function creates a filter Device object
//		and attaches it to the specified target.
//
// Arguments:
//		Pointer to filter Driver object
//		Pointer to the target Device object
//		Pointer to buffer description structure
//
// Return Value:
//		NTSTATUS signaling success or failure
//--
static NTSTATUS
YyAttachFilter(
	IN PDRIVER_OBJECT FilterDriver,
	IN PDEVICE_OBJECT TargetDevice,
	IN PXX_BUFFER_SIZE_INFO BufferInfo
	)
{
	PDEVICE_OBJECT FilterDevice;
    PDEVICE_EXTENSION FilterExtension;

	ULONG TargetMethod;
	NTSTATUS status;

	//
	// Create a Device object without any name...
	//
	status = IoCreateDevice(
				FilterDriver,
				sizeof( DEVICE_EXTENSION ),
				NULL,
				FILE_DEVICE_UNKNOWN,
				0,
				TRUE,
				&FilterDevice );

	if( !NT_SUCCESS( status ))
	{
		return status;
    }

	//
	// Attach to lower-level device.
	//
	status = IoAttachDeviceByPointer(
					FilterDevice,
					TargetDevice );

	if( !NT_SUCCESS( status ))
	{
    	KdPrint( ("[YY DriverEntry] Couldn't attach to the device\n") );
		IoDeleteDevice( FilterDevice );
		return status;
	}

	//
	// Set up filter's Device Extension...
	//
	FilterExtension = FilterDevice->DeviceExtension;
	FilterExtension->DeviceObject = FilterDevice;
	FilterExtension->TargetDevice = TargetDevice;

	FilterExtension->
		BufferInfo.MaxWriteLength = 
							BufferInfo->MaxWriteLength;
	FilterExtension->
		BufferInfo.MaxReadLength = 
							BufferInfo->MaxReadLength;

 	//
	// Copy various items from the target Device object
	//
	FilterDevice->DeviceType = 
					TargetDevice->DeviceType;

	FilterDevice->Characteristics = 
					TargetDevice->Characteristics;

	//
	// Select the same buffering strategy as the
	// lower-level Device object.
	//
	TargetMethod = TargetDevice->Flags &
					(DO_BUFFERED_IO | DO_DIRECT_IO);

	FilterDevice->Flags |= TargetMethod;

	return STATUS_SUCCESS;
}

//++
// Function:
//		YyGetBufferLimits
//
// Description:
//		This function makes an IOCTL call to XXDRIVER
//		to retrieve the max input and output buffer
//		sizes that it supports.
//
// Arguments:
//		Pointer to the target Device object
//		Pointer to buffer description structure
//
// Return Value:
//		(None)
//
//--
static VOID
YyGetBufferLimits(
	IN PDEVICE_OBJECT TargetDevice,
	IN OUT PXX_BUFFER_SIZE_INFO BufferInfo
	)
{
	KEVENT IoctlComplete;
	IO_STATUS_BLOCK Iosb;
	PIRP Irp;
	NTSTATUS status;

	KeInitializeEvent( 
			&IoctlComplete, 
			NotificationEvent, 
			FALSE );

	Irp = IoBuildDeviceIoControlRequest(
				IOCTL_XX_GET_MAX_BUFFER_SIZE,
				TargetDevice,
				NULL,
				0,
				BufferInfo,
				sizeof( XX_BUFFER_SIZE_INFO ),
				FALSE,
				&IoctlComplete,
				&Iosb );

	IoCallDriver( TargetDevice, Irp );

	KeWaitForSingleObject( 
			&IoctlComplete,
			Executive,
			KernelMode,
			FALSE,
			NULL );
}
