//++
// File Name:
//		dispatch.c
//
// Contents:
//		Dispatch routines for driver
//--

//
// Driver-specific header files
//
#include "xxdriver.h"

//
// Forward declarations of local routines
//
// (None)

//
// If the compiler can handle it, save space
// by making various routines pageable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, XxDispatchOpen  )
#pragma alloc_text( page, XxDispatchClose )
#pragma alloc_text( page, XxDispatchRead  )
#pragma alloc_text( page, XxDispatchWrite )
#pragma alloc_text( page, XxDispatchDeviceIoControl )
#endif //ALLOC_PRAGMA

//++
// Function:
//		XxDispatchOpen
//
// Description:
//		This function dispatches CreateFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchOpen(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	KdPrint(( "[XxDispatchOpen] IRP: %8x\n" ));

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

//++
// Function:
//		XxDispatchClose
//
// Description:
//		This function dispatches CloseHandle
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchClose(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	KdPrint(( "[XxDispatchClose] IRP: %8x\n" ));

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

//++
// Function:
//		XxDispatchRead
//
// Description:
//		This function dispatches ReadFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchRead(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION IrpStack = 
				IoGetCurrentIrpStackLocation( Irp );

	ULONG BytesRequested =
			IrpStack->Parameters.Read.Length;

	KdPrint(( "[XxDispatchRead] IRP: %8x (%d bytes)\n",
		Irp, BytesRequested ));

	//
	// Check for zero-length transfers
	//
	if( BytesRequested == 0 ) {
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}

	//
	// Start device operation
	//
	IoMarkIrpPending( Irp );
	IoStartPacket( pDO, Irp, 0, NULL );
	return STATUS_PENDING;
}

//++
// Function:
//		XxDispatchWrite
//
// Description:
//		This function dispatches WriteFile
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchWrite(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION IrpStack = 
				IoGetCurrentIrpStackLocation( Irp );

	ULONG BytesRequested =
			IrpStack->Parameters.Write.Length;

	KdPrint(( "[XxDispatchWrite] IRP: %8x (%d bytes)\n",
		Irp, BytesRequested ));

	//
	// Check for zero-length transfers
	//
	if( IrpStack->Parameters.Write.Length == 0 ) {
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}
	
	//
	// See if output buffer is too large. (This is a
	// purely fabricated restriction -- just to show
	// how a filter driver might get around the
	// problem.
	//
	if( IrpStack->Parameters.Write.Length
		> XX_MAX_BUFFER_SIZE )
	{
		Irp->IoStatus.Status = 
				STATUS_INVALID_BUFFER_SIZE;
		Irp->IoStatus.Information = 0;

		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return STATUS_INVALID_BUFFER_SIZE;
	}

	//
	// Start device operation
	//
	IoMarkIrpPending( Irp );
	IoStartPacket( pDO, Irp, 0, NULL );
	return STATUS_PENDING;
}


//++
// Function:
//		XxDispatchDeviceIoControl
//
// Description:
//		This function dispatches DeviceIoControl
//		requests from Win32
//
// Arguments:
//		Pointer to Device object
//		Pointer to IRP for this request
//
// Return Value:
//		This function returns STATUS_XXX
//--
NTSTATUS
XxDispatchDeviceIoControl(
	IN PDEVICE_OBJECT pDO,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );

	ULONG IoControlCode = 
				IrpStack->Parameters.
								DeviceIoControl.
									IoControlCode;

	PXX_BUFFER_SIZE_INFO InfoBuffer =
			(PXX_BUFFER_SIZE_INFO)
				Irp->AssociatedIrp.SystemBuffer;

	KdPrint(( "[XxDispatchDeviceIoControl] IRP: %8x; IOCTL: %x\n",
		Irp, IoGetFunctionCodeFromCtlCode( IoControlCode )));

	switch( IoControlCode )
	{
		//
		// Dummy functions. Send to Start I/O...
		//
		case IOCTL_XX_FUNCTION1:
		case IOCTL_XX_FUNCTION2:

			IoMarkIrpPending( Irp );
			IoStartPacket( pDO, Irp, 0, NULL );
			return STATUS_PENDING;

		//
		// Return buffer-size limits...
		//
		case IOCTL_XX_GET_MAX_BUFFER_SIZE:

			//
			// Make sure there's enough room
			// in the caller's buffer for the
			// information.
			//
			if( IrpStack->
					Parameters.
						DeviceIoControl.
							OutputBufferLength
				< sizeof( XX_BUFFER_SIZE_INFO ))

			{
				Irp->IoStatus.Status = 
						STATUS_INVALID_BUFFER_SIZE;
				Irp->IoStatus.Information = 0;

				IoCompleteRequest( Irp, IO_NO_INCREMENT );
				return STATUS_INVALID_BUFFER_SIZE;
			}

			//
			// Return size-info...
			//
			InfoBuffer->MaxWriteLength = 
							XX_MAX_BUFFER_SIZE;

			InfoBuffer->MaxReadLength = 
							XX_NO_BUFFER_LIMIT;

			Irp->IoStatus.Information = 
					sizeof( XX_BUFFER_SIZE_INFO );
			
 			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoCompleteRequest( Irp, IO_NO_INCREMENT );
			return STATUS_SUCCESS;

		//
		// Unknown IOCTL code...
		//
		default:
			Irp->IoStatus.Status = 
						STATUS_INVALID_DEVICE_REQUEST;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest( Irp, IO_NO_INCREMENT );
			return STATUS_INVALID_DEVICE_REQUEST;
	}
}

