//++
// File Name:
//		init.c
//
// Contents:
//		Driver initialization code
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
// (None)

//
// If the compiler can handle it, make the DriverEntry
// routine discardable, so that it doesn't waste space
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( init, DriverEntry )
#endif // ALLOC_PRAGMA

//++
// Function:
//		DriverEntry
//
// Description:
//		This function initializes the driver, and
//		creates various NT objects needed to process
//		I/O requests.
//
// Arguments:
//		Pointer to the Driver object
//		Registry path string for driver service key
//
// Return Value:
//		NTSTATUS signaling success or failure
//--
NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	)
{
	NTSTATUS status;

    PDEVICE_EXTENSION pDevExt;
 
	PDEVICE_OBJECT pDevObj1;
	UNICODE_STRING deviceName1;
	UNICODE_STRING linkName1;

	PDEVICE_OBJECT pDevObj2;
	UNICODE_STRING deviceName2;
	UNICODE_STRING linkName2;

	//
	// Export other driver entry points...
	//
 	DriverObject->DriverUnload  = XxDriverUnload;
	DriverObject->DriverStartIo = XxStartIo;
	
	DriverObject->
		MajorFunction[ IRP_MJ_CREATE	] =
			XxDispatchOpen;

	DriverObject->
		MajorFunction[ IRP_MJ_CLOSE	] =
			XxDispatchClose;

	DriverObject->
		MajorFunction[ IRP_MJ_READ	] =
			XxDispatchRead;

	DriverObject->
		MajorFunction[ IRP_MJ_WRITE	] =
			XxDispatchWrite;

	DriverObject->
		MajorFunction[ IRP_MJ_DEVICE_CONTROL ] =
			XxDispatchDeviceIoControl;

	//
   	// Form the NT device names...
   	//
	RtlInitUnicodeString(
		&deviceName1,
		XX_NT_DEVICE_NAME1 );

	RtlInitUnicodeString(
		&deviceName2,
		XX_NT_DEVICE_NAME2 );

    //
    // Create Device objects...
    //
	status = IoCreateDevice(
				DriverObject,
				sizeof( DEVICE_EXTENSION ),
				&deviceName1,
				FILE_DEVICE_UNKNOWN,
				0,
				TRUE,
				&pDevObj1 );

	if( !NT_SUCCESS( status ))
	{
		return status;
    }

	status = IoCreateDevice(
				DriverObject,
				sizeof( DEVICE_EXTENSION ),
				&deviceName2,
				FILE_DEVICE_UNKNOWN,
				0,
				TRUE,
				&pDevObj2 );

	if( !NT_SUCCESS( status ))
	{
		IoDeleteDevice( pDevObj1 );
		return status;
    }

 	//
	// Choose a buffering mechanism
	//
	pDevObj1->Flags |= DO_BUFFERED_IO;
	pDevObj2->Flags |= DO_BUFFERED_IO;

	//
    // Initialize the Device Extension structure
    //
	pDevExt = pDevObj1->DeviceExtension;
    pDevExt->DeviceObject = pDevObj1;

	pDevExt->IrpDelayInterval =
		RtlConvertLongToLargeInteger(
			XX_IRP_DELAY_INTERVAL * 10000 * -1 );

 	KeInitializeTimer( 
		&pDevExt->IrpDelayTimer );

	KeInitializeDpc(
		&pDevExt->IrpDelayDpc,
		XxIrpDelayTimerDpc,
		(PVOID)pDevObj1 );

	//
	// Repeat for second Device Extension
	//
	pDevExt = pDevObj2->DeviceExtension;
    pDevExt->DeviceObject = pDevObj2;

	pDevExt->IrpDelayInterval =
		RtlConvertLongToLargeInteger(
			XX_IRP_DELAY_INTERVAL * 10000 * -1 );

	KeInitializeTimer( 
		&pDevExt->IrpDelayTimer );

	KeInitializeDpc(
		&pDevExt->IrpDelayDpc,
		XxIrpDelayTimerDpc,
		(PVOID)pDevObj2 );

	//
   	// Form the Win32 symbolic link names.
   	//
	RtlInitUnicodeString( 
		&linkName1, 
		XX_WIN32_DEVICE_NAME1 );
		
	RtlInitUnicodeString( 
		&linkName2, 
		XX_WIN32_DEVICE_NAME2 );
		
	//        
	// Create a symbolic links so devices are
	// visible to Win32...
	//
 	status = IoCreateSymbolicLink(
 				&linkName1, &deviceName1 );
    
	//
    // See if the symbolic link was created...
    //
	if( !NT_SUCCESS( status ))
	{
    	IoDeleteDevice( pDevObj2 );
    	IoDeleteDevice( pDevObj1 );
		return status;
    }

	status = IoCreateSymbolicLink(
 				&linkName2, &deviceName2 );
    
	//
    // See if the symbolic link was created...
    //
	if( !NT_SUCCESS( status ))
	{
 		IoDeleteSymbolicLink( &linkName1 );
	   	IoDeleteDevice( pDevObj2 );
    	IoDeleteDevice( pDevObj1 );
    }
	return status;
}

