//++
// File Name:
//		devqueue.c
//
// Contents:
//		Routines to manage the Device Queue object
//		used for input requests
//--

//
// Driver-specific header files...
//
#include "precomp.h"

//
// Forward declarations of local functions
//
// (None)

//++
// Function:
//		XxAlternateStartPacket
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual data transfer.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		(None)
//
// Synchronization:
//		This routine is called at PASSIVE_LEVEL IRQL,
//		but it pops up to DISPATCH_LEVEL.
//
//--
VOID
XxAlternateStartPacket(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PDRIVER_CANCEL CancelFunction
	)
{
	KIRQL OldIrql;

	PDEVICE_EXTENSION DeviceExtension = 
				DeviceObject->DeviceExtension;

	IoAcquireCancelSpinLock( &OldIrql );
	IoSetCancelRoutine( Irp, CancelFunction );

	if( KeInsertDeviceQueue(
			&DeviceExtension->AlternateIrpQueue,
			&Irp->Tail.Overlay.DeviceQueueEntry ))
	{
		IoReleaseCancelSpinLock( OldIrql );
	}
	else
	{
		DeviceExtension->CurrentAlternateIrp = Irp;
		IoReleaseCancelSpinLock( OldIrql );

		KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );
		XxAlternateStartIo( DeviceObject, Irp );
		KeLowerIrql( OldIrql );
	}
}

//++
// Function:
//		XxAlternateStartNextPacket
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual data transfer.
//
// Arguments:
//		Pointer to the Device object
//		BOOLEAN indicating whether 
//				to grab Cancel spin lock
//
// Return Value:
//		(None)
//
// Synchronization:
//		This routine *MUST* be called at 
//		DISPATCH_LEVEL IRQL.
//
//--
VOID
XxAlternateStartNextPacket(
	IN PDEVICE_OBJECT DeviceObject,
	IN BOOLEAN Cancelable
	)
{
	PDEVICE_EXTENSION DeviceExtension = 
				DeviceObject->DeviceExtension;

	PKDEVICE_QUEUE_ENTRY QueueEntry;
	PIRP Irp;
	KIRQL OldIrql;

	if( Cancelable ) 
		IoAcquireCancelSpinLock( &OldIrql );
	
	QueueEntry = 
		KeRemoveDeviceQueue( 
			&DeviceExtension->AlternateIrpQueue );

	if( QueueEntry != NULL )
	{
		Irp = CONTAINING_RECORD(
				QueueEntry,
				IRP,
				Tail.Overlay.DeviceQueueEntry );

		DeviceExtension->CurrentAlternateIrp = Irp;

		if( Cancelable ) 
			IoReleaseCancelSpinLock( OldIrql );

		XxAlternateStartIo( DeviceObject, Irp );
	}
	else 
	{
		DeviceExtension->CurrentAlternateIrp = NULL;

		if( Cancelable ) 
			IoReleaseCancelSpinLock( OldIrql );
	}
}


	
	
	