// File Name:
//		hardware.c
//
// Contents:
//		Routines to hide various hardware details
//

#include "precomp.h"

//
// Forward declarations of local routines
//
// (None)

//
// This is a table of valid baud-rate divisor
// values. It *assumes* the UART's clock is
// running at 1.8432 MHz; a real driver would
// need to do something a little smarter.
//
static ULONG BaudRateDivisors[] = {
	0x0180,		// XX_BAUD_RATE_300
	0x00c0,		// XX_BAUD_RATE_600
	0x0060,		// XX_BAUD_RATE_1200
	0x0030,		// XX_BAUD_RATE_2400
	0x0018,		// XX_BAUD_RATE_4800
	0x000c,		// XX_BAUD_RATE_9600
	0x0006		// XX_BAUD_RATE_19200
};


//++
// Function:
//		XxSetBaudRate
//
// Description:
//		This function sets the baud-rate divisor; 
//		when it runs, it must be synchronized with
//		the ISR.
//
// Arguments:
//		A pointer to a NEW_BAUD_RATE struct
//
// Return Value:
//		TRUE if the change was made
//		FALSE if the baud-rate was invalid
//--
BOOLEAN
XxSetBaudRate(
	PNEW_BAUD_RATE NewBaudRate
	)
{
	UCHAR LineControl;
	PDEVICE_EXTENSION pDE;
	ULONG Divisor;

	//
	// Make sure the chosen baud-rate is
	// within range
	//
	if( (NewBaudRate->Speed < 0 ) ||
		(NewBaudRate->Speed > XX_HIGHEST_BAUD_RATE ))
			return FALSE;

	//
	// Pluck useful values from the context block
	//
	pDE = NewBaudRate->DeviceExtension;
	Divisor = BaudRateDivisors[ NewBaudRate->Speed ];

	//
	// Save the state of the LCR
	//
	LineControl = XxReadLineControl( pDE );

	//
	// Turn on the DLAB bit and put the new
	// divisor value in place
	//
	XxWriteLineControl( 
		pDE,
		(UCHAR)(LineControl | XX_LCR_DLAB));

	XxWriteBaudMsb( 
		pDE,
		(UCHAR)((Divisor & 0xff00) >> 8));

	XxWriteBaudLsb( pDE, (UCHAR)(Divisor & 0xff));

	//
	// Restore the state of the LCR
	//
	XxWriteLineControl( pDE, LineControl );

	return TRUE;
}

//++
// Function:
//		XxInitializeHardware
//
// Description:
//		This function puts the hardware into a known
//		state, whih includes disabling all UART
//		interrupts.
//
// Arguments:
//		A pointer to a Device Extension
//
// Return Value:
//		Always TRUE
//
// Synchronization:
//		This function must be synchronized with
//		the Interrupt Service routine.
//
//--
BOOLEAN
XxInitializeHardware(
	PDEVICE_EXTENSION pDE
	)
{
	NEW_BAUD_RATE BaudRate;
	ULONG i;

	//
	// Disable all UART interrupts
	//
 	XxWriteIntEnable( pDE, (UCHAR)0 );
	XxWriteModemControl( pDE, (UCHAR)0 );

	//
	// Set up the FIFO Control Register
	//
	//
	// It's necessary to disable the FIFOs, read
	// from the input data register, then set the
	// FIFO-state. This is a work-around for an
	// odd hardware problem with some chips.
	//
	XxWriteFifoControl( pDE, 0 );
	XxReadDataBuffer( pDE );
	XxWriteFifoControl(
		pDE,
		(UCHAR) (XX_FCR_ENABLE |
				 XX_FCR_RCVR_RESET |
				 XX_FCR_TXMT_RESET |
				 pDE->InputFifoTriggerLevel) );

	//
	// Set up the Line Control Register
	//
	XxWriteLineControl( 
		pDE,
		(UCHAR) (pDE->DataBits |
				 pDE->StopBits |
				 pDE->Parity ));

	//
	// Now read the input data register until
	// the receive-data-ready flag drops. Stop
	// after five iterations.
	//
	for( i=0; i<5; i++ )
	{
		XxReadDataBuffer( pDE );
		if(( XxReadLineStatus( pDE ) & 
					XX_LSR_DATA_RDY ) == 0 )
		{
			break;
		}
	}

	//
	// Next, read the low bits of the Modem Status
	// register until they drop.
	//
	for( i=0; i<5; i++ )
	{
		if(( XxReadModemStatus( pDE ) &
				( XX_MSR_DELTA_CTS |
				  XX_MSR_DELTA_DSR |
				  XX_MSR_DELTA_RI  |
				  XX_MSR_DELTA_DCD )) == 0 )
		{
			break;
		}
	}

	//
	// Select a default baud rate
	//
	BaudRate.Speed = XX_BAUD_RATE_9600;
	BaudRate.DeviceExtension = pDE;
	XxSetBaudRate( &BaudRate );

	return TRUE;
}

//++
// Function:
//		XxEnableInterrupts
//
// Description:
//		This function turns on UART interupts.
//		It expects to be synchronized with the
//		ISR.
//
// Arguments:
//		A pointer to a Device Extension
//
// Return Value:
//		Always TRUE
//--
BOOLEAN
XxEnableInterrupts(
	PDEVICE_EXTENSION pDE
	)
{
	UCHAR Control;
	ULONG i;

	//
	// Turn on the interesting interrupts
	//
	XxWriteIntEnable( pDE, XX_IEN_RDR | XX_IEN_TBE );

	//
	// Read the Interrupt ID register a few times
	// until the no-interrupt flag pops up.
	//
	for( i=0; i<5; i++ )
	{
		if(( XxReadIntId( pDE ) &
			 XX_IIR_NO_INTERRUPT ) == 1 )
		{
			break;
		}
	}

	//
	// Set the OUT2 bit of the Modem Control
	// register. This is what really enables
	// interrupts.
	//
	Control = XxReadModemControl( pDE );
	Control |= XX_MCR_OUT2;
	XxWriteModemControl( pDE, Control ); 

	return TRUE;
}

//++
// Function:
//		XxDisableInterrupts
//
// Description:
//		This function turns UART interupts off.
//		It expects to be synchronized with the
//		ISR.
//
// Arguments:
//		A pointer to a Device Extension
//
// Return Value:
//		Always TRUE
//--
BOOLEAN
XxDisableInterrupts(
	PDEVICE_EXTENSION pDE
	)
{
	UCHAR Control;

	Control = XxReadModemControl( pDE );
	Control &= ~XX_MCR_OUT2;
	XxWriteModemControl( pDE, Control ); 

	XxWriteIntEnable( pDE, (UCHAR)0 );

	return TRUE;
}
