//++
// File Name:
//		input.c
//
// Contents:
//		Routines to process UART input requests
//--

//
// Driver-specific header files...
//
#include "precomp.h"

//
// Forward declarations of local functions
//
// (None)

//++
// Function:
//		XxAlternateStartIo
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual data transfer.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		(None)
//--
VOID
XxAlternateStartIo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	KIRQL OldIrql;

	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );
		
	PDEVICE_EXTENSION DeviceExtension =
				DeviceObject->DeviceExtension;

 	//
	// Before starting the operation, see if the
	// Cancel routine has run between the time
	// the IRP was removed from the Device Queue
	// and now. Start by grabbing the Cancel spin
	// lock.
	//
	IoAcquireCancelSpinLock( &OldIrql );

	//
	// Check the state of the Cancel flag. If 
	// it's set, release the spin lock and return
	// immediately. The IRP has already been  
	// processed by the Cancel routine.
	//
	if( Irp->Cancel )
	{
		IoReleaseCancelSpinLock( OldIrql );
		return;
	}
	//
	// If the Cancel flag is clear, remove the
	// IRP from the cancelable state and start
	// processing it.
	//
	else
	{
		IoSetCancelRoutine( Irp, NULL );
		IoReleaseCancelSpinLock( OldIrql );
	}
		
	switch( IrpStack->MajorFunction ) {
		
		case IRP_MJ_READ:
			//
			// Set up counts and byte pointer
			//
			DeviceExtension->InputBytesRequested = 
				IrpStack->Parameters.Read.Length;

			DeviceExtension->InputBytesRemaining = 
				DeviceExtension->InputBytesRequested;

			DeviceExtension->pInputBuffer =
				Irp->AssociatedIrp.SystemBuffer;

			//
			// Try to start receiving bytes. If
			// there's a problem, use the DPC
			// routine to fail the IRP.
			//
			if( !KeSynchronizeExecution(
					DeviceExtension->pInterrupt,
					XxReceiveBytes,
					DeviceExtension ))
			{
				XxDpcForInputs(
					NULL,
					DeviceObject,
					Irp,
					DeviceExtension );
			}
			break;
		//
		// Should never get here -- just get rid
		// of the packet...
		//	
		default:

			//
			// Fail the IRP and start the next one.
			//
			Irp->IoStatus.Status =
						STATUS_NOT_SUPPORTED;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest(
				Irp,
				IO_NO_INCREMENT );

			XxAlternateStartNextPacket( 
							DeviceObject, 
							TRUE );
			break;
	}
}

//++
// Function:
//		XxReceiveBytes
//
// Description:
//		This function does whatever is 
//		necessary to prepare for input.
//
// Arguments:
//		Pointer to the Device Extension
//
// Return Value:
//		TRUE
//
// Synchronization:
//		This routine expects to be synchronized 
//		with the ISR, running at DIRQL holding 
//		the Interrupt spin lock.
//
//--
BOOLEAN
XxReceiveBytes( 
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION DeviceExtension = Context;

	//
	// Setting this flag is enough to get
	// ISR to process input data.
	//
	DeviceExtension->InputInterruptsValid = TRUE;

	return TRUE;
}

//++
// Function:
//		XxDpcForInputs
//
// Description:
//		This function performs the low-IRQL
//		post-processing of input requests
//
// Arguments:
//		Pointer to a DPC object
//		Pointer to the Device object
//		Pointer to the IRP for this request
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxDpcForInputs(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION DeviceExtension = Context;

	//
	// Determine number of bytes transferred
	//
	Irp->IoStatus.Information =
			DeviceExtension->InputBytesRequested -
			DeviceExtension->InputBytesRemaining;

	//
	// Figure out what the final status
	// should be
	//
	Irp->IoStatus.Status = STATUS_SUCCESS;

	//
	// If we're being called directly from Start I/O,
	// don't give the user any priority boost.
	//
	if( Dpc == NULL )
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	else
		IoCompleteRequest( Irp, IO_SERIAL_INCREMENT );

	//
	// This one's done. Begin working on the next
	//
	XxAlternateStartNextPacket( DeviceObject, TRUE );
}



	
	
	