//++
// File Name:
//		isr.c
//
// Contents:
//		Routines to handle UART interrupts
//--

//
// Driver-specific header files...
//
#include "precomp.h"

//
// Forward declarations of local functions
//
static VOID
XxHandleInputFifoTrigger(
	IN PDEVICE_EXTENSION pDE
	);

static VOID
XxHandleInputFifoTimeOut(
	IN PDEVICE_EXTENSION pDE
	);

//++
// Function:
//		XxIsr
//
// Description:
//		This function is the main interrupt
//		handler for the UART
//
// Arguments:
//		Pointer to the Interrupt object
//		Pointer to the Device Extension 
//
// Return Value:
//		TRUE if we handled the interrupt
//
// Synchronization:
//		Called by the system at DIRQL, holding
//		the Interrupt spin lock for the UART.
//
//--
BOOLEAN
XxIsr(
	IN PKINTERRUPT Interrupt,
	IN PVOID ServiceContext
	)
{
	PDEVICE_EXTENSION pDE = ServiceContext;

	PDEVICE_OBJECT pDevice = pDE->DeviceObject;

	UCHAR InterruptId = XxReadIntId( pDE );

	//
	// See if this device requested an interrupt
	//
	if(( InterruptId & XX_IIR_NO_INTERRUPT ) != 0 )
		return FALSE;

	//
	// It did. See what kind it was.
	//
	do
	{
		InterruptId &= XX_IIR_INTERRUPT_ID_MASK;

		switch( InterruptId )
		{
			//
			// For the moment, just ignore
			// any errors
			//
			case XX_IIR_ERR:
				XxReadLineStatus( pDE );
				break;

			//
			// Input-FIFO hit its trigger-level;
			// get some more characters.
			//
			case XX_IIR_RDA:
				XxHandleInputFifoTrigger( pDE );
				break;

			//
			// Input-FIFO has timed out while
			// holding data; empty it and complete
			// the IRP
			//
			case XX_IIR_FIFO_TMO:
				XxHandleInputFifoTimeOut( pDE );
				break;

			//
			// If we're really doing output,
			// refill the transmit FIFO...
			//
			case XX_IIR_TBE:
				if( pDE->OutputInterruptsValid )
					if( !XxTransmitBytes( pDE ))
							IoRequestDpc( 
								pDevice, 
								pDevice->CurrentIrp,
								(PVOID)pDE );
				break;

			//
			// Ignore modem events
			//
			case XX_IIR_RS232:
				XxReadModemStatus( pDE );
				break;
		}

		InterruptId = XxReadIntId( pDE );
	} while(( InterruptId & 
				XX_IIR_NO_INTERRUPT ) == 0 );

	return TRUE;
}


//
// FIFO trigger-level handler
//
static VOID
XxHandleInputFifoTrigger(
	IN PDEVICE_EXTENSION pDE
	)
{
	ULONG i;

	//
	// Read one less than the number of bytes in
	// the FIFO; this guarantees a FIFO time-out
	// which will end the read request.
	//
	for( i=0; i < pDE->InputFifoSize - 1; i++ )
	{
		if( pDE->InputInterruptsValid &&
			pDE->InputBytesRemaining > 0 )
		{
			*pDE->pInputBuffer++ =
				XxReadDataBuffer( pDE );
			pDE->InputBytesRemaining--;
		}
		else XxReadDataBuffer( pDE );
	}
}

//
// FIFO time-out handler
//
static VOID
XxHandleInputFifoTimeOut(
	IN PDEVICE_EXTENSION pDE
	)
{
	//
	// Keep reading bytes until the FIFO is empty
	// If there's room, put them into the system
	// buffer; otherwise, drop them on the floor
	//
	while( XxReadLineStatus( pDE ) & XX_LSR_DATA_RDY )
	{
		if( pDE->InputInterruptsValid &&
			pDE->InputBytesRemaining > 0 )
		{
			*pDE->pInputBuffer++ =
				XxReadDataBuffer( pDE );
			pDE->InputBytesRemaining--;
		}
		else XxReadDataBuffer( pDE );
	}

	//
	// If this was a spurious interrupt, we're done.
	// If it was for real, then complete the current
	// read-request.
	//
	if( pDE->InputInterruptsValid )
	{
		pDE->InputInterruptsValid = FALSE;

		KeInsertQueueDpc(
			&pDE->AlternateDpc,
			(PVOID)pDE->CurrentAlternateIrp,
			(PVOID)pDE );
	}
}