//++
// File Name:
//		crash.c
//
// Contents:
//		Functions that help the driver crash
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Global variables
//
static ULONG CrashPoint;	// Determines where the
							// driver will crash.

//
// Forward declarations of local functions
//
// (None)

//
// If the compiler can handle it, save space
// by making various routines pageable
//    
#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, XxGetCrashPoint )
#endif //ALLOC_PRAGMA

//++
// Function:
//		XxGetCrashPoint
//
// Description:
//		This function tries to pull the value 
//		of the CrashPoint variable from the
//		driver's Parameters subkey in the Registry.
//		If anything goes wrong, it uses the default
//		value.
//
// Arguments:
//		(None)
//
// Return Value:
//		(None)
//--
VOID
XxGetCrashPoint(
	VOID
	)
{
	RTL_QUERY_REGISTRY_TABLE QueryTable[2];

	//
	// Fabricate a Registry query.
	//
	RtlZeroMemory( QueryTable, sizeof( QueryTable ));

	QueryTable[0].Name	= L"CrashPoint";
	QueryTable[0].Flags	= RTL_QUERY_REGISTRY_DIRECT;
	QueryTable[0].EntryContext = &CrashPoint;

	//
	// Look for the CrashPoint value 
	// in the Registry.
	//
	if( !NT_SUCCESS( 
			RtlQueryRegistryValues(
					RTL_REGISTRY_SERVICES,
					XX_DRIVER_NAME
						L"\\Parameters",
					QueryTable,
					NULL, NULL )))
	{
		//
		// Something went wrong; use default
		//
		CrashPoint = XX_DEFAULT_CRASH_TYPE;

		KdPrint((
			"Using default value for CrashPoint\n" ));
	}

	KdPrint(( "CrashPoint: %d\n", CrashPoint ));
}

//++
// Function:
//		XxTryToCrash
//
// Description:
//		If this function is called from the place
//		in the driver where a crash should occur,
//		generate an access violation. Otherwise,
//		the call is a no-op. 
//
// Arguments:
//		Indicator of place in driver where this
//		function is being called.
//
// Return Value:
//		(None)
//--
VOID
XxTryToCrash(
	IN ULONG CurrentLocation
	)
{
	UCHAR c;

	//
	// If this is the place to crash, generate
	// an access violation
	//
	if( CurrentLocation == CrashPoint )
	{
		KdPrint(( "About to generate crash...\n" ));
		c = *(PUCHAR)0;
	}
}

//++
// Function:
//		XxBugCheckCallback
//
// Description:
//		If this function is called by the system
//		when a crash occurs.
//
// Arguments:
//		Address of our crash buffer
//		Size of the crash buffer
//
// Return Value:
//		(None)
//--
VOID
XxBugCheckCallback(
	IN PVOID Buffer,
	IN ULONG Length
	)
{
	PCRASH_BLOCK pCrashData = (PCRASH_BLOCK)Buffer;

	//
	// Store crucial driver data
	//
	pCrashData->Stuff			= CrashPoint;
	pCrashData->MoreStuff		= 0x2222;
	pCrashData->EvenMoreStuff	= 0x3333;
}