//++
// File Name:
//		transfer.c
//
// Contents:
//		Data transfer routines for crash driver
//--

//
// Driver-specific header files...
//
#include "xxdriver.h"

//
// Forward declarations of local functions
//
// (None)

//++
// Function:
//		XxStartIo
//
// Description:
//		This function is responsible for doing any
//		function-code specific pre-processing and 
//		then starting the actual data transfer.
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the IRP for this request
//
// Return Value:
//		(None)
//--
VOID
XxStartIo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PIO_STACK_LOCATION	IrpStack = 
		IoGetCurrentIrpStackLocation( Irp );
		
	PDEVICE_EXTENSION pDE =
		DeviceObject->DeviceExtension;

	KdPrint(( "Entering Start I/O routine...\n" ));
		
	//
	// See if we should croak...
	//
	XxTryToCrash( XX_CRASH_START_IO );

	switch( IrpStack->MajorFunction ) {
		
		case IRP_MJ_WRITE:
			//
			// Set up counts and byte pointer
			//
			pDE->BytesRequested = 
				IrpStack->Parameters.Write.Length;

			pDE->BytesRemaining = 
				pDE->BytesRequested;

			pDE->pBuffer =
				Irp->AssociatedIrp.SystemBuffer;

			//
			// Use DPC routine to complete the
			// request.
			//
			KeInsertQueueDpc(
				&pDE->CompletionDPC,
				Irp,
				(PVOID)pDE );

		/*	XxDpcForIsr(
					NULL,
					DeviceObject,
					Irp,
					pDE ); */
			break;
		//
		// Should never get here -- just get rid
		// of the packet...
		//	
		default:
			Irp->IoStatus.Status =
						STATUS_NOT_SUPPORTED;
			Irp->IoStatus.Information = 0;
			IoCompleteRequest(
				Irp,
				IO_NO_INCREMENT );
			IoStartNextPacket( DeviceObject, FALSE );
			break;
	}
}


//++
// Function:
//		XxCompleteRequest
//
// Description:
//		This function performs the low-IRQL
//		post-processing of I/O requests
//
// Arguments:
//		Pointer to a DPC object
//		Pointer to the Device object
//		Pointer to the IRP for this request
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxCompleteRequest(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION pDE = Context;

	KdPrint(( "Entering DPC routine...\n" ));

	//
	// Calculate transfer length
	//
	Irp->IoStatus.Information =
			pDE->BytesRequested -
			pDE->BytesRemaining;

	Irp->IoStatus.Status = STATUS_SUCCESS;

	//
	// See if we should croak...
	//
	XxTryToCrash( XX_CRASH_DPC );

	//
	// This one's done. Begin working on the next
	//
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	IoStartNextPacket( DeviceObject, FALSE );
}

//++
// Function:
//		XxTimer
//
// Description:
//		This function gets called once each
//		second
//
// Arguments:
//		Pointer to the Device object
//		Pointer to the Device Extension
//
// Return Value:
//		(None)
//--
VOID
XxTimer(
	IN PDEVICE_OBJECT DeviceObject,
	IN PVOID Context
	)
{
	PDEVICE_EXTENSION pDE = Context;

	if( --pDE->TimeoutCounter <= 0 )
	{
		KdPrint(( "Timer has expired...\n" ));
		IoStopTimer( DeviceObject );

		//
		// See if we should croak...
		//
		XxTryToCrash( XX_CRASH_TIMER );
	}
}



