//++
// File Name:
//		eventlog.c
//
// Contents:
//		Event Log management functions for
//		the data collection DLL
//--

//
// All-inclusive header file
//
#include "xxperf.h"

//
// Data global to this module
//
static HANDLE hEventLog		= NULL;
static DWORD  dwUserCount	= 0;
static DWORD  dwLogLevel	= DEFAULT_LOG_LEVEL;

//
// Forward declarations of local routines
//
// (None)


//++
// Function:
//		XxOpenEventLog
//
// Description:
//		This function opens a connection with 
//		the Event Log. If it's already open,
//		just increment a counter. 
//
// Arguments:
//		(None)
//
// Return Value:
//		(None)
//--
VOID
XxOpenEventLog(
	VOID
	)
{
    HKEY hLogLevelKey;
	
	TCHAR LogLevelKeyName[] = 
			"SYSTEM\\CurrentControlSet\\Services\\"
			XX_DRIVER_NAME 
			"\\Parameters";

	TCHAR LogLevelValueName[] = "EventLogLevel";

    LONG status;

    DWORD dwLevelFromRegistry;
    DWORD dwValueType;
    DWORD dwValueSize;
   
	if( hEventLog == NULL )
	{
		//
		// Find the logging-level value in the
		// Registry. If it's not there, or we
		// have problems with the Registry, just
		// use the default value.
		//
		status = RegOpenKeyEx(
					HKEY_LOCAL_MACHINE,
					LogLevelKeyName,
					0,
					KEY_READ,
					&hLogLevelKey );

	   if( status == ERROR_SUCCESS)
	   { 
			dwValueSize = 
				sizeof( dwLevelFromRegistry );

			status = RegQueryValueEx(
						hLogLevelKey,
						LogLevelValueName,
						NULL,
						&dwValueType,
						(LPBYTE)&dwLevelFromRegistry,
						&dwValueSize );

		   if( status == ERROR_SUCCESS )
		   {
			   dwLogLevel = dwLevelFromRegistry;
		   }
		   else
		   {
			   dwLogLevel = DEFAULT_LOG_LEVEL;
		   }
		   
		   RegCloseKey( hLogLevelKey );
	   }
	   else
	   {
		   dwLogLevel = DEFAULT_LOG_LEVEL;
	   }

	   //
	   // Open a handle to the Event Log
	   //
	   hEventLog = RegisterEventSource(
						(LPTSTR)NULL,
						XX_EVENT_LOG_KEY );

	   if( hEventLog != NULL )
	   {
			XxLogInformation(
				LOG_LEVEL_DEBUG_MAX,
				XXPERF_LOG_OPEN );
	   }
	}

    //
	// Increment count of Event Log users
	//
    if( hEventLog != NULL ) dwUserCount++;
}


//++
// Function:
//		XxCloseEventLog
//
// Description:
//		This function breaks our connection with
//		the Event Log once everyone is finished
//		using it.
//
// Arguments:
//		(None)
//
// Return Value:
//		(None)
//--
VOID
XxCloseEventLog(
	VOID
	)

{
	if( hEventLog != NULL )
	{
		dwUserCount--;
        if( dwUserCount <= 0 )
		{ 
			XxLogInformation(
				LOG_LEVEL_DEBUG_MAX,
				XXPERF_CLOSING_LOG );
			DeregisterEventSource( hEventLog );
			hEventLog = NULL;
		}
	}
}

//++
// Function:
//		XxLogInformation
//
// Description:
//		This function logs an informational
//		message without any dump data.
//
// Arguments:
//		LOG_LEVEL of the message
//		Message code number
//
// Return Value:
//		(None)
//--
VOID
XxLogInformation(
	IN DWORD dwMessageLevel,
	IN DWORD dwMessage
	)
{
	if( ( hEventLog != NULL ) &&
		( dwLogLevel > LOG_LEVEL_NONE ) &&
		( dwMessageLevel <= dwLogLevel ))
	{
		ReportEvent(
			hEventLog,
			EVENTLOG_INFORMATION_TYPE,
			0,
			dwMessage,
			(PSID)NULL,
			0,
			0,
			NULL,
			NULL );
	}
}


//++
// Function:
//		XxLogError
//
// Description:
//		This function logs an error
//		message without any dump data.
//
// Arguments:
//		LOG_LEVEL of the message
//		Message code number
//
// Return Value:
//		(None)
//--
VOID
XxLogError(
	IN DWORD dwMessageLevel,
	IN DWORD dwMessage
	)
{
	if( ( hEventLog != NULL ) &&
		( dwLogLevel > LOG_LEVEL_NONE ) &&
		( dwMessageLevel <= dwLogLevel ))
	{
		ReportEvent(
			hEventLog,			// Event Log handle
			EVENTLOG_ERROR_TYPE,// Message type
			0,					// Category
			dwMessage,			// Message code
			(PSID)NULL,			// Security ID
			0,					// Number of insertion strings
			0,					// Byte-count of dump data
			NULL,				// Array of insertion strings
			NULL );				// Ptr to dump-data or NULL
	}
}

//++
// Function:
//		XxLogErrorWithData
//
// Description:
//		This function logs an error
//		message that has dump data.
//
// Arguments:
//		LOG_LEVEL of the message
//		Message code number
//		Pointer to an array of DWORD dump data
//		Count of bytes in dump data
//
// Return Value:
//		(None)
//--
VOID
XxLogErrorWithData(
	IN DWORD	dwMessageLevel,
	IN DWORD	dwMessage,
	IN LPDWORD	lpdwDumpData,
	IN DWORD	cbDumpDataSize
	)
{
	if( ( hEventLog != NULL ) &&
		( dwLogLevel > LOG_LEVEL_NONE ) &&
		( dwMessageLevel <= dwLogLevel ))
	{
		ReportEvent(
			hEventLog,
			EVENTLOG_ERROR_TYPE,
			0,
			dwMessage,
			(PSID)NULL,
			0,
			cbDumpDataSize,
			NULL,
			(PVOID)lpdwDumpData );
	}
}
