//++
// File Name:
//		xxperf.c
//
// Contents:
//		Data collection DLL that interfaces with
//		the performance-monitoring system
//--

//
// All-inclusive header file
//
#include "xxperf.h"

//
// Data global to this module
//
static HANDLE hDevice;
static DWORD dwOpenCount = 0;
static BOOL bInitialized = FALSE;

//
// Initialized object header defined
// in data.c
//
extern XX_HEADER_DEFINITION XxObjectHeader;

//
// Forward declarations of routines
//
PM_OPEN_PROC	XxPerfOpen;
PM_COLLECT_PROC	XxPerfCollect;
PM_CLOSE_PROC	XxPerfClose;


//++
// Function:
//		XxPerfOpen
//
// Description:
//		This function sets up the data collection
//		DLL for later calls. 
//
// Arguments:
//		Unicode string describing objects to be opened
//
// Return Value:
//		ERROR_SUCCESS if everything works
//		ERROR_XXX if something goes wrong
//--
DWORD
XxPerfOpen(
    LPWSTR lpDeviceNames
    )
{
    HKEY	hKeyDriverPerf;
	DWORD	dwFirstCounter;
	DWORD	dwFirstHelp;
    DWORD	dwType;
    DWORD	dwSize;
    DWORD	dwStatus;

	//
	// DLL must keep track of how many it is opened
	// because the Registry routines may call it in
	// response to multiple performance queries.
	//
	if( dwOpenCount == 0 )
	{
		//
		// Errors in the DLL will go to the Event
		// Log. Open it here.
		XxOpenEventLog();

		//
		// Open a handle to the device generating
		// the performance data. Save it in a global
		// variable where the rest of the DLL can get
		// to it.
		//
		hDevice = CreateFile(
					XX_WIN32_DEVICE_NAME,
					GENERIC_READ,
					FILE_SHARE_READ |
						FILE_SHARE_WRITE,
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL,
					NULL );
		if( hDevice == INVALID_HANDLE_VALUE )
		{
			dwStatus = GetLastError();
 			XxLogErrorWithData(
				LOG_LEVEL_NORMAL,
				XXPERF_CANT_OPEN_DEVICE_HANDLE,
				&dwStatus, sizeof( dwStatus ) );
			XxCloseEventLog();
			return dwStatus;
		}

		//
		// Open the Performance subkey of the driver's
		// service key in the Registry.
		//
		dwStatus = RegOpenKeyEx(
						HKEY_LOCAL_MACHINE,
						"SYSTEM\\CurrentControlSet"
							"\\Services\\XxDriver"
								"\\Performance",
						0L,
						KEY_ALL_ACCESS,
						&hKeyDriverPerf );

		if( dwStatus != ERROR_SUCCESS )
		{
			XxLogErrorWithData(
				LOG_LEVEL_NORMAL,
				XXPERF_CANT_OPEN_DRIVER_KEY,
				&dwStatus, sizeof( dwStatus ));

			CloseHandle( hDevice );
			XxCloseEventLog();
			return dwStatus;
		}

		//
		// Get base index of first object or counter
		//
		dwSize = sizeof (DWORD);
		dwStatus = RegQueryValueEx(
					hKeyDriverPerf, 
					"First Counter",
					0L,
					&dwType,
					(LPBYTE)&dwFirstCounter,
					&dwSize);

		if( dwStatus != ERROR_SUCCESS )
		{
			XxLogErrorWithData(
				LOG_LEVEL_NORMAL,
				XXPERF_CANT_READ_FIRST_COUNTER,
				&dwStatus, sizeof( dwStatus ));

			RegCloseKey( hKeyDriverPerf );
			CloseHandle( hDevice );
			XxCloseEventLog();
			return dwStatus;
		}

		//
		// Get base index of first help text
		//
		dwSize = sizeof (DWORD);
		dwStatus = RegQueryValueEx(
					hKeyDriverPerf,
					"First Help",
					0L,
					&dwType,
					(LPBYTE)&dwFirstHelp,
					&dwSize );

		if( dwStatus != ERROR_SUCCESS )
		{
			XxLogErrorWithData(
				LOG_LEVEL_NORMAL,
				XXPERF_CANT_READ_FIRST_HELP,
				&dwStatus, sizeof( dwStatus ));

			RegCloseKey( hKeyDriverPerf );
			CloseHandle( hDevice );
			XxCloseEventLog();
			return dwStatus;
		}

		//
		// Don't need Registy handle anymore
		//
		RegCloseKey( hKeyDriverPerf );
 
		//
		// Prepare the  static portions of header
		//
		//
		// Initialize PERF_OBJECT_TYPE struct
		//
		XxObjectHeader.XxDevice.
				ObjectNameTitleIndex =
						dwFirstCounter + XXDEVICE;

		XxObjectHeader.XxDevice.
					ObjectHelpTitleIndex = 
						dwFirstHelp + XXDEVICE;

		//
		// Initialize 1st PERF_COUNTER_DEFINITION
		//
		XxObjectHeader.Interrupts.
					CounterNameTitleIndex = 
						dwFirstCounter + INTERRUPTS;
		XxObjectHeader.Interrupts.
					CounterHelpTitleIndex = 
						dwFirstHelp + INTERRUPTS;

		//
		// Initialize 2nd PERF_COUNTER_DEFINITION
		//
		XxObjectHeader.Operations.
					CounterNameTitleIndex = 
						dwFirstCounter + OPERATIONS;
		XxObjectHeader.Operations.
					CounterHelpTitleIndex = 
						dwFirstHelp + OPERATIONS;

		//
		// Mark DLL as successfully initialized
		//
		bInitialized = TRUE;
	}

	//
	// One way or another, there's one more
	// thread using the DLL.
	//
	dwOpenCount++;

	return ERROR_SUCCESS;
}

//++
// Function:
//		XxPerfCollect
//
// Description:
//		This function collects one sample 
//		of data for the specified objects.
//
// Arguments:
//		Unicode string describing objects to be opened
//		Ptr to data buffer
//		Size of buffer
//		Number of separate objects in buffer
//	
//
// Return Value:
//		ERROR_MORE_DATA if buffer too small
//		ERROR_SUCCESS all other situations
//--
DWORD
XxPerfCollect(
    IN LPWSTR lpValueName,
    IN OUT LPVOID *lppData,
    IN OUT LPDWORD lpcbTotalBytes,
    IN OUT LPDWORD lpNumObjectTypes
	)
{
	DWORD dwQueryType;
	DWORD dwStatus;
	DWORD dwBytesReturned;
    PPERF_COUNTER_BLOCK pPerfCounterBlock;
	PXX_HEADER_DEFINITION pXxObjectHeader;
 
    //
    // Make sure the XxPerfOpen function
	// ran successfully.
    //
    if( !bInitialized )
	{
		//
		// Initialization failed. Quit.
		//
		*lpcbTotalBytes = (DWORD) 0;
		*lpNumObjectTypes = (DWORD) 0;
		return ERROR_SUCCESS;
	}

    //
	// Parse argument string to see what kind
	// of data the caller is asking for.
    //
	dwQueryType = XxGetPerfQueryType( lpValueName );
    
    if( dwQueryType == PERF_QUERY_TYPE_FOREIGN )
	{
		//
		// Can't service foreign requests.
		//
		*lpcbTotalBytes = (DWORD) 0;
		*lpNumObjectTypes = (DWORD) 0;
		return ERROR_SUCCESS;
	}

	if( dwQueryType == PERF_QUERY_TYPE_ITEMS )
	{
		//
		// Caller passed a list of object-type
		// indexes. See if we provide data for any
		// of them.
		//
		if( !XxIsNumberInList( 
				XxObjectHeader.
					XxDevice.
						ObjectNameTitleIndex,
				lpValueName ))
		{
			//
			// We don't provide data for any of the
			// objects they've asked for. Quit.
			//
			*lpcbTotalBytes = (DWORD) 0;
			*lpNumObjectTypes = (DWORD) 0;
			return ERROR_SUCCESS;
		}
	}

	//
	// Make sure there's enough room in the caller's
	// buffer for everything we're sending back.
	//
	if( *lpcbTotalBytes < 
			( sizeof( XX_HEADER_DEFINITION ) +
				sizeof( XX_PERF_DATA )))
	{
		*lpcbTotalBytes = (DWORD) 0;
		*lpNumObjectTypes = (DWORD) 0;
		return ERROR_MORE_DATA;
	}

    //
	// Copy the pre-initialized PERF_OBJECT_TYPE and
	// PERF_COUNTER_DEFINITIONs into the caller's data
	// buffer.
	//
	pXxObjectHeader = 
			(PXX_HEADER_DEFINITION) *lppData;

	memmove(
		pXxObjectHeader,
		&XxObjectHeader,
		sizeof( XX_HEADER_DEFINITION ));

	//
	// Get a pointer to the PERF_COUNTER_BLOCK in
	// the caller's buffer.
	//
	pPerfCounterBlock = 
		(PPERF_COUNTER_BLOCK) &pXxObjectHeader[1];

	//
	// Call the kernel-mode driver to dump raw
	// counter data into the caller's buffer. This
	// IOCTL also fills in the ByteLength field of
	// the PERF_COUNTER_BLOCK.
	//
	if( !DeviceIoControl(
			hDevice,
			IOCTL_XX_GET_PERF_DATA,
			NULL,
			0,
			(LPVOID)pPerfCounterBlock,
			sizeof( XX_PERF_DATA ),
			&dwBytesReturned,
			NULL ))
	{
		dwStatus = GetLastError();
		XxLogErrorWithData(
			LOG_LEVEL_NORMAL,
			XXPERF_CANT_READ_PERF_DATA,
			&dwStatus, sizeof( dwStatus ) );

		*lpcbTotalBytes = (DWORD) 0;
		*lpNumObjectTypes = (DWORD) 0;
		return ERROR_SUCCESS;
	}

	//
    // Update caller's arguments
	//    
	*lppData = 
		(PUCHAR)pPerfCounterBlock +
				sizeof( XX_PERF_DATA );

	*lpNumObjectTypes = XX_NUM_OBJECT_TYPES;

	*lpcbTotalBytes =
		sizeof( XX_HEADER_DEFINITION ) +
			sizeof( XX_PERF_DATA );

	return ERROR_SUCCESS;
}

//++
// Function:
//		XxPerfClose
//
// Description:
//		This function closes down the
//		collection DLL. 
//
// Arguments:
//		(None)
//	
//
// Return Value:
//		ERROR_SUCCESS always
//--
DWORD
XxPerfClose( )
{
	//
	// One less thread is using the DLL
	//
	if( --dwOpenCount <= 0 )
	{
		//
		// When the last thread goes,
		// shut everything down
		//
		CloseHandle( hDevice );

		XxCloseEventLog();	
	}	
	return ERROR_SUCCESS;
}

